#!/usr/bin/env pmpython
#
# Copyright (c) 2023 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-arguments,too-many-lines, bad-continuation
# pylint: disable=redefined-outer-name,unnecessary-lambda
#
import signal
import sys
import time
from pcp import pmapi, pmcc
from cpmapi import PM_CONTEXT_ARCHIVE
from cpmapi import PM_CONTEXT_HOST

# By default we are assuming we are running on the latest pcp version
# Later will figure out the version and will change the newVersionFlag

newVersionFlag = True

SYS_METRICS = ["kernel.uname.sysname", "kernel.uname.release",
               "kernel.uname.nodename", "kernel.uname.machine", "hinv.ncpu"]

ZONESTAT_METRICS = ["mem.zoneinfo.managed", "mem.zoneinfo.nr_mlock", "mem.zoneinfo.present",
                    "mem.zoneinfo.scanned", "mem.zoneinfo.nr_unstable", "mem.zoneinfo.nr_page_table_pages",
                    "mem.zoneinfo.nr_shmem", "mem.zoneinfo.nr_free_pages", "mem.zoneinfo.nr_active_file",
                    "mem.zoneinfo.nr_dirty", "mem.zoneinfo.nr_writeback", "mem.zoneinfo.free",
                    "mem.zoneinfo.nr_unevictable", "mem.zoneinfo.nr_alloc_batch", "mem.zoneinfo.low",
                    "mem.zoneinfo.nr_slab_reclaimable", "mem.zoneinfo.nr_kernel_stack", "mem.zoneinfo.numa_interleave",
                    "mem.zoneinfo.workingset_nodereclaim", "mem.zoneinfo.nr_isolated_anon", "mem.zoneinfo.nr_bounce",
                    "mem.zoneinfo.numa_other", "mem.zoneinfo.nr_file_pages", "mem.zoneinfo.numa_hit",
                    "mem.zoneinfo.nr_isolated_file", "mem.zoneinfo.nr_anon_transparent_hugepages",
                    "mem.zoneinfo.nr_inactive_file", "mem.zoneinfo.spanned", "mem.zoneinfo.nr_written",
                    "mem.zoneinfo.numa_foreign", "mem.zoneinfo.nr_vmscan_write", "mem.zoneinfo.nr_free_cma",
                    "mem.zoneinfo.nr_writeback_temp", "mem.zoneinfo.nr_slab_unreclaimable",
                    "mem.zoneinfo.nr_vmscan_immediate_reclaim", "mem.zoneinfo.nr_mapped", "mem.zoneinfo.nr_anon_pages",
                    "mem.zoneinfo.high", "mem.zoneinfo.protection", "mem.zoneinfo.numa_local", "mem.zoneinfo.numa_miss",
                    "mem.zoneinfo.nr_active_anon", "mem.zoneinfo.workingset_refault", "mem.zoneinfo.nr_dirtied",
                    "mem.zoneinfo.workingset_activate", "mem.zoneinfo.nr_inactive_anon", "mem.zoneinfo.min"]

ZONESTAT_NEW_METRICS = ["mem.zoneinfo.nr_foll_pin_acquired", "mem.zoneinfo.nr_foll_pin_released",
                        "mem.zoneinfo.nr_zspages", "mem.zoneinfo.workingset_restore_anon",
                        "mem.zoneinfo.workingset_refault_anon", "mem.zoneinfo.nr_zone_active_anon",
                        "mem.zoneinfo.nr_zone_inactive_file", "mem.zoneinfo.nr_swapcached",
                        "mem.zoneinfo.nr_file_pmdmapped", "mem.zoneinfo.workingset_active_file",
                        "mem.zoneinfo.workingset_refault_file", "mem.zoneinfo.cma",
                        "mem.zoneinfo.workingset_active_anon", "mem.zoneinfo.nr_zone_write_pending",
                        "mem.zoneinfo.nr_shmem_pmdmapped", "mem.zoneinfo.nr_shmem_hugepages",
                        "mem.zoneinfo.nr_kernel_misc_reclaimable", "mem.zoneinfo.nr_zone_active_file",
                        "mem.zoneinfo.nr_file_hugepages", "mem.zoneinfo.nr_zone_unevictable",
                        "mem.zoneinfo.nr_zone_inactive_anon", "mem.zoneinfo.workingset_restore_file"]

ZONEINFO_PER_NODE = {
    "nr_inactive_anon"              :   "mem.zoneinfo.nr_inactive_anon",
    "nr_active_anon"                :   "mem.zoneinfo.nr_active_anon",
    "nr_inactive_file"              :   "mem.zoneinfo.nr_inactive_file",
    "nr_active_file"                :   "mem.zoneinfo.nr_active_file",
    "nr_unevictable"                :   "mem.zoneinfo.nr_unevictable",
    "nr_slab_reclaimable"           :   "mem.zoneinfo.nr_slab_reclaimable",
    "nr_slab_unreclaimable"         :   "mem.zoneinfo.nr_slab_unreclaimable",
    "nr_isolated_anon"              :   "mem.zoneinfo.nr_isolated_anon",
    "nr_isolated_file"              :   "mem.zoneinfo.nr_isolated_file",
    "nr_anon_pages"                 :   "mem.zoneinfo.nr_anon_pages",
    "nr_mapped"                     :   "mem.zoneinfo.nr_mapped",
    "nr_file_pages"                 :   "mem.zoneinfo.nr_file_pages",
    "nr_dirty"                      :   "mem.zoneinfo.nr_dirty",
    "nr_writeback"                  :   "mem.zoneinfo.nr_writeback",
    "nr_writeback_temp"             :   "mem.zoneinfo.nr_writeback_temp",
    "nr_shmem"                      :   "mem.zoneinfo.nr_shmem",
    "nr_anon_transparent_hugepages" :   "mem.zoneinfo.nr_anon_transparent_hugepages",
    "nr_unstable"                   :   "mem.zoneinfo.nr_unstable",
    "nr_vmscan_write"               :   "mem.zoneinfo.nr_vmscan_write",
    "nr_vmscan_immediate_reclaim"   :   "mem.zoneinfo.nr_vmscan_immediate_reclaim",
    "nr_dirtied"                    :   "mem.zoneinfo.nr_dirtied",
    "nr_written"                    :   "mem.zoneinfo.nr_written",
    "nr_shmem_hugepages"            :   "mem.zoneinfo.nr_shmem_hugepages",
    "nr_shmem_pmdmapped"            :   "mem.zoneinfo.nr_shmem_pmdmapped",
    "nr_file_hugepages"             :   "mem.zoneinfo.nr_file_hugepages",
    "nr_file_pmdmapped"             :   "mem.zoneinfo.nr_file_pmdmapped",
    "nr_kernel_misc_reclaimable"    :   "mem.zoneinfo.nr_kernel_misc_reclaimable"
}

ZONEINFO_PAGE_INFO = [
        ("pages free"   ,   "mem.zoneinfo.free"),
        ("      min"    ,   "mem.zoneinfo.min"),
        ("      low"    ,   "mem.zoneinfo.low"),
        ("      high"   ,   "mem.zoneinfo.high"),
        ("      spanned",   "mem.zoneinfo.spanned"),
        ("      present",   "mem.zoneinfo.present"),
        ("      managed",   "mem.zoneinfo.managed")
]

ZONEINFO_NUMBER_ZONE    =   {
    "nr_free_pages"         :   "mem.zoneinfo.nr_free_pages",
    "nr_mlock"              :   "mem.zoneinfo.nr_mlock",
    "nr_page_table_pages"   :   "mem.zoneinfo.nr_page_table_pages",
    "nr_kernel_stack"       :   "mem.zoneinfo.nr_kernel_stack",
    "nr_bounce"             :   "mem.zoneinfo.nr_bounce",
    "nr_free_cma"           :   "mem.zoneinfo.nr_free_cma",
    "numa_hit"              :   "mem.zoneinfo.numa_hit",
    "numa_miss"             :   "mem.zoneinfo.numa_miss",
    "numa_foreign"          :   "mem.zoneinfo.numa_foreign",
    "numa_interleave"       :   "mem.zoneinfo.numa_interleave",
    "numa_other"            :   "mem.zoneinfo.numa_other",
    "nr_zone_active_anon"   :   "mem.zoneinfo.nr_zone_active_anon",
    "nr_zone_inactive_file" :    "mem.zoneinfo.nr_zone_inactive_file",
    "nr_zone_active_file"   :   "mem.zoneinfo.nr_zone_active_file",
    "nr_zone_unevictable"   :   "mem.zoneinfo.nr_zone_unevictable",
    "nr_zone_write_pending" :   "mem.zoneinfo.nr_zone_write_pending",
    "nr_zspages"            :   "mem.zoneinfo.nr_zspages",
    "numa_local"            :   "mem.zoneinfo.numa_local",
    "nr_zone_inactive_anon" :   "mem.zoneinfo.nr_zone_inactive_anon"
}


class ReportingMetricRepository:

    def __init__(self, group):
        self.group = group
        self.current_cached_values = {}

    def __sorted(self, data):
        return dict(sorted(data.items(), key=lambda item: item[0].lower()))

    def __fetch_current_value(self, metric):
        val = dict(map(lambda x: (x[1], x[2]), self.group[metric].netValues))
        val = self.__sorted(val)
        return dict(val)

    def current_value(self, metric):
        if metric not in self.group:
            return None
        if self.current_cached_values.get(metric) is None:
            first_value = self.__fetch_current_value(metric)
            self.current_cached_values[metric] = first_value
        return self.current_cached_values[metric]


class ZoneStatUtil:
    def __init__(self, metrics_repository, all_metrics_repository):
        self.__metric_repository = metrics_repository
        self.__all_metrics_repository = all_metrics_repository
        self.report = ReportingMetricRepository(self.__metric_repository)

    def names(self):
        data = self.report.current_value('mem.zoneinfo.present')
        return sorted(data.keys())

    def metric_value(self, metric, node):
        if metric in self.__all_metrics_repository:
            data = self.report.current_value(metric)
            data = data.get(node, "0")
            if data != "0":
                data = data/4
            return int(data)
        return None

    # creating the list of protection values for particular node
    def protection(self, node):
        data = self.report.current_value("mem.zoneinfo.protection")
        return [value // 4 for key, value in data.items() if key.startswith(node)]

    def protection_names(self, node_name):
        data = self.report.current_value("mem.zoneinfo.protection")
        return {
            key.split("::" + node_name)[0] + "::" + node_name
            for key in data.keys()
            if node_name in key
        } or None


class ZoneinfoReport(pmcc.MetricGroupPrinter):
    def __init__(self, opts, group):
        self.opts = opts
        self.group = group
        self.samples = opts.samples
        self.context = opts.context

    def __get_ncpu(self, group):
        return group['hinv.ncpu'].netValues[0][2]

    def __print_machine_info(self, context):
        timestamp = self.group.pmLocaltime(context.timestamp.tv_sec)
        # Please check strftime(3) for different formatting options.
        # Also check TZ and LC_TIME environment variables for more
        # information on how to override the default formatting of
        # the date display in the header
        time_string = time.strftime("%x", timestamp.struct_time())
        header_string = ''
        header_string += context['kernel.uname.sysname'].netValues[0][2] + '  '
        header_string += context['kernel.uname.release'].netValues[0][2] + '  '
        header_string += '(' + context['kernel.uname.nodename'].netValues[0][2] + ')  '
        header_string += time_string + '  '
        header_string += context['kernel.uname.machine'].netValues[0][2] + '  '
        print("%s  (%s CPU)" % (header_string, self.__get_ncpu(context)))

    # Function to format the node name as per the /proc/zoneinfo naming convention
    def __format_node_name(self, input_str):
        parts = input_str.split("::")
        if len(parts) != 2 or not parts[1].startswith("node"):
            return "Invalid input format"
        node_num = parts[1][4:]
        return "Node {}, zone    {}".format(node_num, parts[0])

    def __print_old_version_values(self, manager):
        try:
            total_nodes = manager.names()
            total_nodes = sorted(total_nodes, key=lambda
                                 x: ((x.split('::')[0] != 'DMA')*2,  # Move 'DMA' entries to the front
                                     x.split('::')[0]))
            for node in total_nodes:
                print(self.__format_node_name(str(node)))
                print("\tper-node status")
                # print(ZONEINFO_PER_NODE)
                for key, value in ZONEINFO_PER_NODE.items():
                    print("\t{} {}".format(key, manager.metric_value(value, node)))
                for key, value in ZONEINFO_PAGE_INFO:
                    print("\t{} {}".format(key, manager.metric_value(value, node)))
                print(" " * 14 + "protection " + str([int(i) for i in sorted(manager.protection(node))]))
                for key, value in ZONEINFO_NUMBER_ZONE.items():
                    print("\t{} {}".format(key, manager.metric_value(value, node)))
        except IndexError:
            print("Got some error while printing values for old version zoneinfo")

    def __print_values(self, manager):
        total_nodes = manager.names()
        node_names = {key.split('::')[1] for key in total_nodes}
        # sort the node names in decreasing order
        node_names = sorted(node_names, key=lambda x: int(x[4:]) if x.startswith('node') else float('inf'))
        try:
            # lopping through all the available nodes
            for node_name in node_names:
                # get all the types available for the current node on which i'm
                node_types = manager.protection_names(node_name)
                if node_types is None:
                    return
                nodes = set(node_types).intersection(total_nodes)
                nodes = sorted(nodes, key=lambda x: ((x.split('::')[0] != 'DMA')*2,  # Move 'DMA' entries to the front
                                                     x.split('::')[0]))
                print("NODE {:>2}, per-node status".format(node_name[4:]))
                for key, value in ZONEINFO_PER_NODE.items():
                    print("\t{} {}".format(key, manager.metric_value(value, node_name)))
                for node in nodes:
                    print(self.__format_node_name(node))
                    for key, value in ZONEINFO_PAGE_INFO:
                        print("\t{} {}".format(key, manager.metric_value(value, node)))
                    print(" " * 14 + "protection " + str([int(i) for i in manager.protection(node)]))
                    for key, value in ZONEINFO_NUMBER_ZONE.items():
                        print("\t{} {}".format(key, manager.metric_value(value, node)))
                # finding the remaining type of nodes data which i haven't printed so far
                # these nodes will have only pages information for them so just printing them out
                remaining_nodes = set(node_types) - set(nodes)
                if remaining_nodes:
                    for node in sorted(remaining_nodes):
                        print(self.__format_node_name(node))
                        for key, value in ZONEINFO_PAGE_INFO:
                            print("\t", key, manager.metric_value(value, node))
                        print(" " * 14 + "protection " + str([int(i) for i in manager.protection(node)]))
                else:
                    continue
        except IndexError:
            print("Got some error while printing values for zoneinfo")

    def print_report(self, group, timestamp, manager_zoneinfo, manager_all):
        def __print_zone_status():
            zonestatus = ZoneStatUtil(manager_zoneinfo, manager_all)
            if zonestatus.names():
                try:
                    self.__print_machine_info(group)
                    print("TimeStamp = {}".format(timestamp))
                    if newVersionFlag:
                        self.__print_values(zonestatus)
                    else:
                        self.__print_old_version_values(zonestatus)
                except IndexError:
                    print("Incorrect machine info due to some missing metrics")

        if self.context != PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_zone_status()
            sys.exit(0)
        elif self.context == PM_CONTEXT_ARCHIVE and self.samples is None:
            __print_zone_status()
        elif self.samples >= 1:
            __print_zone_status()
            self.samples -= 1

    def report(self, manager):
        group = manager["sysinfo"]
        self.samples = self.opts.pmGetOptionSamples()
        t_s = group.contextCache.pmLocaltime(int(group.timestamp))
        timestamp = time.strftime(ZoneinfoOptions.timefmt, t_s.struct_time())
        self.print_report(group, timestamp, manager['zoneinfo'], manager['allinfo'])


class ZoneinfoOptions(pmapi.pmOptions):
    timefmt = "%H:%M:%S"

    def __init__(self):
        pmapi.pmOptions.__init__(self, "a:s:Z:zV?")
        self.pmSetLongOptionHeader("General options")
        self.pmSetLongOptionHostZone()
        self.pmSetLongOptionTimeZone()
        self.pmSetLongOptionHelp()
        self.pmSetLongOptionSamples()
        self.pmSetLongOptionVersion()
        self.samples = None
        self.context = None


def updateZoneinfoDataToBeFetched():
    ZoneinfoNumberZoneElementsToBeRemoved = {
        "nr_zone_active_anon"           :   "mem.zoneinfo.nr_zone_active_anon",
        "nr_zone_inactive_file"         :    "mem.zoneinfo.nr_zone_inactive_file",
        "nr_zone_active_file"           :   "mem.zoneinfo.nr_zone_active_file",
        "nr_zone_unevictable"           :   "mem.zoneinfo.nr_zone_unevictable",
        "nr_zone_write_pending"         :   "mem.zoneinfo.nr_zone_write_pending",
        "nr_zspages"                    :   "mem.zoneinfo.nr_zspages",
        "numa_local"                    :   "mem.zoneinfo.numa_local",
        "nr_zone_inactive_anon"         :   "mem.zoneinfo.nr_zone_inactive_anon"
    }

    ZoneinfoPerNodeElementsToBeRemoved = {
        "nr_shmem_hugepages"            :   "mem.zoneinfo.nr_shmem_hugepages",
        "nr_shmem_pmdmapped"            :   "mem.zoneinfo.nr_shmem_pmdmapped",
        "nr_file_hugepages"             :   "mem.zoneinfo.nr_file_hugepages",
        "nr_file_pmdmapped"             :   "mem.zoneinfo.nr_file_pmdmapped",
        "nr_kernel_misc_reclaimable"    :   "mem.zoneinfo.nr_kernel_misc_reclaimable"
    }

    for key in ZoneinfoNumberZoneElementsToBeRemoved:
        del ZONEINFO_NUMBER_ZONE[key]

    for key in ZoneinfoPerNodeElementsToBeRemoved:
        del ZONEINFO_PER_NODE[key]


if __name__ == '__main__':
    try:
        opts = ZoneinfoOptions()
        mngr = pmcc.MetricGroupManager.builder(opts, sys.argv)
        opts.context = mngr.type
        ZONESTAT_METRICS = ZONESTAT_METRICS + ZONESTAT_NEW_METRICS

        if opts.context == PM_CONTEXT_HOST:
            context = pmapi.pmContext(opts.context)
        elif opts.context == PM_CONTEXT_ARCHIVE:
            context = pmapi.pmContext.fromOptions(opts,sys.argv)
        try:
            metric_ids = context.pmLookupName(ZONESTAT_METRICS)
            descs= context.pmLookupDescs(metric_ids)
        except pmapi.pmErr as error:
            newVersionFlag = False
            updateZoneinfoDataToBeFetched()
            ZONESTAT_METRICS = [x for x in ZONESTAT_METRICS if x not in ZONESTAT_NEW_METRICS]

        ALL_METRICS = ZONESTAT_METRICS + SYS_METRICS
        mngr["zoneinfo"] = ZONESTAT_METRICS
        mngr["sysinfo"] = SYS_METRICS
        mngr["allinfo"] = ALL_METRICS
        mngr.printer = ZoneinfoReport(opts, mngr)
        sts = mngr.run()
        sys.exit(sts)
    except pmapi.pmErr as error:
        sys.stderr.write('%s\n' % (error.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except IOError:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env pmpython
#
# Copyright (c) 2023 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-arguments,too-many-lines, bad-continuation
# pylint: disable=redefined-outer-name,unnecessary-lambda
#

import sys
import time
import signal
from pcp import pmapi, pmcc
from cpmapi import PM_CONTEXT_ARCHIVE

METRICS = ["mem.physmem",
            "mem.util.free",
            "mem.util.available",
            "mem.util.bufmem",
            "mem.util.cached",
            "mem.util.swapCached",
            "mem.util.active",
            "mem.util.inactive",
            "mem.util.active_anon",
            "mem.util.inactive_anon",
            "mem.util.active_file",
            "mem.util.inactive_file",
            "mem.util.unevictable",
            "mem.util.mlocked",
            "mem.util.swapTotal",
            "mem.util.swapFree",
            "mem.util.dirty",
            "mem.util.writeback",
            "mem.util.anonpages",
            "mem.util.mapped",
            "mem.util.shared",
            "mem.util.kreclaimable",
            "mem.util.slab",
            "mem.util.slabReclaimable",
            "mem.util.slabUnreclaimable",
            "mem.util.kernelStack",
            "mem.util.pageTables",
            "mem.util.NFS_Unstable",
            "mem.util.bounce",
            "mem.vmstat.nr_writeback_temp",
            "mem.util.commitLimit",
            "mem.util.committed_AS",
            "mem.util.vmallocTotal",
            "mem.util.vmallocUsed",
            "mem.util.vmallocChunk",
            "mem.util.percpu",
            "mem.util.corrupthardware",
            "mem.util.anonhugepages",
            "mem.vmstat.nr_shmem_hugepages",
            "mem.vmstat.nr_shmem_pmdmapped",
            "mem.util.filehugepages",
            "mem.util.filepmdmapped",
            "mem.util.cmatotal",
            "mem.zoneinfo.nr_free_cma",
            "mem.util.hugepagesTotal",
            "mem.util.hugepagesFree",
            "mem.util.hugepagesRsvd",
            "mem.util.hugepagesSurp",
            "hinv.hugepagesize",
            "mem.util.hugetlb",
            "mem.util.directMap4k",
            "mem.util.directMap2M",
            "mem.util.directMap1G"]

METRICS_DESC = ["MemTotal",
                "MemFree",
                "MemAvailable",
                "Buffers",
                "Cached",
                "SwapCached",
                "Active",
                "Inactive",
                "Active(anon)",
                "Inactive(anon)",
                "Active(file)",
                "Inactive(file)",
                "Unevictable",
                "Mlocked",
                "SwapTotal",
                "SwapFree",
                "Dirty",
                "Writeback",
                "AnonPages",
                "Mapped",
                "Shmem",
                "KReclaimable",
                "Slab",
                "SReclaimable",
                "SUnreclaim",
                "KernelStack",
                "PageTables",
                "NFS_Unstable",
                "Bounce",
                "WritebackTmp",
                "CommitLimit",
                "Committed_AS",
                "VmallocTotal",
                "VmallocUsed",
                "VmallocChunk",
                "Percpu",
                "HardwareCorrupted",
                "AnonHugePages",
                "ShmemHugePages",
                "ShmemPmdMapped",
                "FileHugePages",
                "FilePmdMapped",
                "CmaTotal",
                "CmaFree",
                "HugePages_Total_NO_kb",
                "HugePages_Free_NO_kb",
                "HugePages_Rsvd_NO_kb",
                "HugePages_Surp_NO_kb",
                "Hugepagesize",
                "Hugetlb",
                "DirectMap4k",
                "DirectMap2M",
                "DirectMap1G"]

SYS_METRICS = ["kernel.uname.sysname",
                "kernel.uname.release",
                "kernel.uname.nodename",
                "kernel.uname.machine",
                "hinv.ncpu"]

ALL_METRICS = METRICS + SYS_METRICS

class MeminfoReport(pmcc.MetricGroupPrinter):
    def __init__(self, opts):
        self.opts = opts
        self.Machine_info_count = 0

    def __get_ncpu(self, group):
        return group['hinv.ncpu'].netValues[0][2]

    def __print_machine_info(self, group, context):
        timestamp = context.pmLocaltime(group.timestamp.tv_sec)
        # Please check strftime(3) for different formatting options.
        # Also check TZ and LC_TIME environment variables for more
        # information on how to override the default formatting of
        # the date display in the header
        time_string = time.strftime("%x", timestamp.struct_time())
        header_string = ''
        header_string += group['kernel.uname.sysname'].netValues[0][2] + '  '
        header_string += group['kernel.uname.release'].netValues[0][2] + '  '
        header_string += '(' + group['kernel.uname.nodename'].netValues[0][2] + ')  '
        header_string += time_string + '  '
        header_string += group['kernel.uname.machine'].netValues[0][2] + '  '

        print("%s  (%s CPU)" % (header_string, self.__get_ncpu(group)))

    def getMetricName(self, idx):
        metric_name = ""
        units = ""
        if METRICS_DESC[idx][-6:] == "_NO_kb":
            metric_name = METRICS_DESC[idx][:-6]
        elif METRICS_DESC[idx] == "Hugepagesize":
            metric_name = METRICS_DESC[idx]
            units = "B"
        else:
            metric_name = METRICS_DESC[idx]
            units = "kB"
        return metric_name, units

    def report(self, manager):
        group = manager["sysinfo"]
        try:
            if not self.Machine_info_count:
                self.__print_machine_info(group, manager)
                self.Machine_info_count = 1
        except IndexError:
            # missing some metrics
            return

        group = manager["meminfo"]

        self.opts.pmGetOptionSamples()

        t_s = group.contextCache.pmLocaltime(int(group.timestamp))
        time_string = time.strftime(MeminfoOptions.timefmt, t_s.struct_time())
        print("Timestamp".ljust(18) + ": " + time_string)

        idx = 0
        for metric in METRICS:
            try:
                val = group[metric].netValues[0][2]
            except IndexError:
                metric_name, units = self.getMetricName(idx)
                print("%-17s : NA"%(metric_name))

                idx += 1
                continue

            metric_name, units = self.getMetricName(idx)
            if units == "B":
                val = int(val / 1024)
                units = "kB"
            print("%-17s : %s %s"%(metric_name, val, units))

            idx += 1
        print("")

        if MeminfoOptions.context is not PM_CONTEXT_ARCHIVE and self.opts.pmGetOptionSamples() is None:
            sys.exit(0)

class MeminfoOptions(pmapi.pmOptions):
    context = None
    timefmt = "%m/%d/%Y %H:%M:%S"

    def __init__(self):
        pmapi.pmOptions.__init__(self, "a:s:S:T:z:A:t:")
        self.pmSetLongOptionStart()
        self.pmSetLongOptionFinish()
        self.pmSetLongOptionHelp()

if __name__ == '__main__':
    try:
        opts = MeminfoOptions()
        mngr = pmcc.MetricGroupManager.builder(opts,sys.argv)
        MeminfoOptions.context = mngr.type

        missing = mngr.checkMissingMetrics(ALL_METRICS)
        if missing is not None:
            sys.stderr.write('Error: not all required metrics are available\nMissing: %s\n' % (missing))
            sys.exit(1)

        mngr["meminfo"] = METRICS
        mngr["sysinfo"] = SYS_METRICS
        mngr.printer = MeminfoReport(opts)
        sts = mngr.run()
        sys.exit(sts)
    except IOError:
        signal.signal(signal.SIGPIPE,signal.SIG_DFL)
    except pmapi.pmErr as error:
        sys.stderr.write("%s %s\n"%(error.progname(), error.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env pmpython
# -*- coding: utf-8 -*-

# Exercises hardcoded metrics, including conflict handling when such metrics are received

import sys
import socket
import glob
import os

utils_path = os.path.abspath(os.path.join("utils"))
sys.path.append(utils_path)

import pmdastatsd_test_utils as utils

utils.print_test_file_separator()
print(os.path.basename(__file__))

ip = "0.0.0.0"
port = 8125
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)


def run_test():
    utils.print_test_section_separator()
    utils.pmdastatsd_install()
    utils.print_metric('statsd.pmda.received')
    utils.print_metric('statsd.pmda.parsed')
    utils.print_metric('statsd.pmda.aggregated')
    utils.print_metric('statsd.pmda.dropped')
    utils.print_metric('statsd.pmda.time_spent_parsing')
    utils.print_metric('statsd.pmda.time_spent_aggregating')
    utils.print_metric('statsd.pmda.metrics_tracked')
    utils.print_metric('statsd.pmda.settings.duration_aggregation_type')
    utils.print_metric('statsd.pmda.settings.parser_type')
    utils.print_metric('statsd.pmda.settings.port')
    utils.print_metric('statsd.pmda.settings.debug_output_filename')
    utils.print_metric('statsd.pmda.settings.verbose')
    utils.print_metric('statsd.pmda.settings.max_unprocessed_packets')
    utils.print_metric('statsd.pmda.settings.max_udp_packet_size')

    sock.sendto("pmda.received:1|c".encode("utf-8"), (ip, port))
    sock.sendto("pmda.parsed:1|c".encode("utf-8"), (ip, port))
    sock.sendto("pmda.aggregated:1|c".encode("utf-8"), (ip, port))
    sock.sendto("pmda.dropped:1|c".encode("utf-8"), (ip, port))
    sock.sendto("pmda.time_spent_parsing:1|c".encode("utf-8"), (ip, port))
    sock.sendto("pmda.time_spent_aggregating:1|c".encode("utf-8"), (ip, port))
    sock.sendto("pmda.metrics_tracked:1|c".encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.duration_aggregation_type:1|c'.encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.parser_type:1|c'.encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.port:1|c'.encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.debug_output_filename:1|c'.encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.verbose:1|c'.encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.max_unprocessed_packets:1|c'.encode("utf-8"), (ip, port))
    sock.sendto('pmda.settings.max_udp_packet_size:1|c'.encode("utf-8"), (ip, port))

    utils.print_metric('statsd.pmda.received')
    utils.print_metric('statsd.pmda.parsed')
    utils.print_metric('statsd.pmda.aggregated')
    utils.print_metric('statsd.pmda.dropped')
    utils.print_metric('statsd.pmda.metrics_tracked')
    
    utils.pmdastatsd_remove()
    utils.restore_config()

run_test()



#!/usr/bin/pmpython
#
# Copyright (c) 2025 Oracle and/or its affiliates.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=bad-whitespace,too-many-arguments,too-many-lines, bad-continuation, line-too-long
# pylint: disable=redefined-outer-name,unnecessary-lambda, wildcard-import, unused-wildcard-import
#
from datetime import datetime
from pcp import pmapi
from pcp import pmcc
import sys
import time

ROCESTAT_HW_METRICS = [
    'rocestat.hw.link.link_error_recovery',
    'rocestat.hw.link.link_downed',
    'rocestat.hw.link.local_link_integrity_errors',
    'rocestat.hw.rnr.rnr_nak_retry_err',
    'rocestat.hw.resp.resp_local_length_error',
    'rocestat.hw.resp.resp_cqe_error',
    'rocestat.hw.resp.resp_cqe_flush_error',
    'rocestat.hw.resp.resp_remote_access_errors',
    'rocestat.hw.req.req_remote_invalid_request',
    'rocestat.hw.req.req_cqe_error',
    'rocestat.hw.req.req_cqe_flush_error',
    'rocestat.hw.req.duplicate_request',
    'rocestat.hw.req.rx_read_requests',
    'rocestat.hw.req.rx_atomic_requests',
    'rocestat.hw.req.req_remote_access_errors',
    'rocestat.hw.req.rx_write_requests',
    'rocestat.hw.mcast.multicast_rcv_packets',
    'rocestat.hw.mcast.multicast_xmit_packets',
    'rocestat.hw.ucast.unicast_rcv_packets',
    'rocestat.hw.ucast.unicast_xmit_packets',
    'rocestat.hw.rcv.port_rcv_errors',
    'rocestat.hw.rcv.port_rcv_remote_physical_errors',
    'rocestat.hw.rcv.port_rcv_packets',
    'rocestat.hw.rcv.port_rcv_data',
    'rocestat.hw.rcv.port_rcv_constraint_errors',
    'rocestat.hw.rcv.port_rcv_switch_relay_errors',
    'rocestat.hw.xmit.port_xmit_data',
    'rocestat.hw.xmit.port_xmit_constraint_errors',
    'rocestat.hw.xmit.port_xmit_wait',
    'rocestat.hw.xmit.port_xmit_packets',
    'rocestat.hw.xmit.port_xmit_discards',
    'rocestat.hw.roce_slow_restart_trans',
    'rocestat.hw.roce_slow_restart_cnps',
    'rocestat.hw.roce_slow_restart',
    'rocestat.hw.roce_adp_retrans_to',
    'rocestat.hw.clear_counters',
    'rocestat.hw.local_ack_timeout_err',
    'rocestat.hw.lifespan',
    'rocestat.hw.implied_nak_seq_err',
    'rocestat.hw.packet_seq_err',
    'rocestat.hw.roce_adp_retrans',
    'rocestat.hw.out_of_buffer',
    'rocestat.hw.out_of_sequence',
    'rocestat.hw.VL15_dropped',
    'rocestat.hw.excessive_buffer_overrun_errors',
    'rocestat.hw.symbol_error',
    'rocestat.hw.rp_cnp_handled',
    'rocestat.hw.rp_cnp_ignored',
    'rocestat.hw.np_cnp_sent',
    'rocestat.hw.np_ecn_marked_roce_packets',
]

ROCESTAT_HW_METRICS_DESC = [
    'link_error_recovery',
    'link_downed',
    'local_link_integrity_errors',
    'rnr_nak_retry_err',
    'resp_local_length_error',
    'resp_cqe_error',
    'resp_cqe_flush_error',
    'resp_remote_access_errors',
    'req_remote_invalid_request',
    'req_cqe_error',
    'req_cqe_flush_error',
    'duplicate_request',
    'rx_read_requests',
    'rx_atomic_requests',
    'req_remote_access_errors',
    'rx_write_requests',
    'multicast_rcv_packets',
    'multicast_xmit_packets',
    'unicast_rcv_packets',
    'unicast_xmit_packets',
    'port_rcv_errors',
    'port_rcv_remote_physical_errors',
    'port_rcv_packets',
    'port_rcv_data',
    'port_rcv_constraint_errors',
    'port_rcv_switch_relay_errors',
    'port_xmit_data',
    'port_xmit_constraint_errors',
    'port_xmit_wait',
    'port_xmit_packets',
    'port_xmit_discards',
    'roce_slow_restart_trans',
    'roce_slow_restart_cnps',
    'roce_slow_restart',
    'roce_adp_retrans_to',
    'clear_counters',
    'local_ack_timeout_err',
    'lifespan',
    'implied_nak_seq_err',
    'packet_seq_err',
    'roce_adp_retrans',
    'out_of_buffer',
    'out_of_sequence',
    'VL15_dropped',
    'excessive_buffer_overrun_errors',
    'symbol_error',
    'rp_cnp_handled',
    'rp_cnp_ignored',
    'np_cnp_sent',
    'np_ecn_marked_roce_packets',
]

ROCESTAT_PER_LANE_METRICS = ['rocestat.lane.tx_bytes', 'rocestat.lane.rx_bytes', 'rocestat.lane.rx_pause']


class RoceStatOptions(pmapi.pmOptions):
    context = None
    timefmt = "%H:%M:%S"
    samples = 0
    hw_stats_flag = False
    lane_stats_flag = False

    def extra_options(self, opt, optarg, index):
        if opt == "hw_stats":
            RoceStatOptions.hw_stats_flag = True

        elif opt == "lane_stats":
            RoceStatOptions.lane_stats_flag = True

    def __init__(self):
        pmapi.pmOptions.__init__(self, "a:s:S:T:t:")
        self.pmSetLongOption("hw_stats", 0, "", "", "Display hw counters")
        self.pmSetLongOption("lane_stats", 0, "", "", "Display lane stats")
        self.pmSetLongOptionStart()
        self.pmSetLongOptionFinish()
        self.pmSetLongOption("samples", 1, "s", "COUNT",
                             "Number of samples to collect")
        self.pmSetLongOptionArchive()
        self.pmSetLongOptionHelp()

        self.pmSetOptionCallback(self.extra_options)


class ROCeStatReport:
    prev_counter_data = None
    def __init__(self, samples):
        self.samples = samples

    def report(self, manager):
        group = manager["rocestat"]

        # commented out by kenj to fix make check failure
        # 25 Nov 2025
        # opts.pmGetOptionSamples()

        timestamp = manager.pmLocaltime(group.timestamp.tv_sec)
        time_string = time.strftime("%x", timestamp.struct_time()) + " "
        t_s = group.contextCache.pmLocaltime(int(group.timestamp))
        time_string += time.strftime(RoceStatOptions.timefmt, t_s.struct_time())
        print("zzz <%s>" % (time_string))
        hw_stats_flag = RoceStatOptions.hw_stats_flag
        lane_stats_flag = RoceStatOptions.lane_stats_flag

        # HW Counters
        if hw_stats_flag or (hw_stats_flag == lane_stats_flag):
            header = ["Counter"]
            for val in group[ROCESTAT_HW_METRICS[0]].netValues:
                if not val[1].endswith("_delta"):
                    header.append(val[1])
            card_idxs = {}
            for i in range(1, len(header)):
                card_idxs[header[i]] = i
            prev_counter_data = self.prev_counter_data
            curr_counter_data = {}
            curr_counter_data["ts"] = datetime.now()
            for card in card_idxs:
                curr_counter_data[card] = {}

            for metric_name in ROCESTAT_HW_METRICS_DESC:
                metric_idx = ROCESTAT_HW_METRICS_DESC.index(metric_name)
                metric = ROCESTAT_HW_METRICS[metric_idx]
                for val in group[metric].netValues:
                    card_name, value = val[1], val[2]
                    curr_counter_data[card_name][metric_name] = value

            counters = []
            for counter in ROCESTAT_HW_METRICS_DESC:
                row = [counter] + ["" for _ in range(len(card_idxs.keys()))]
                for card, card_data in curr_counter_data.items():
                    if card == "ts":
                        continue
                    card_idx = card_idxs[card]
                    count = card_data[counter]
                    if prev_counter_data is not None and card in prev_counter_data:
                        count_delta = count - prev_counter_data[card][counter]
                        time_diff = (curr_counter_data["ts"] - prev_counter_data["ts"]).total_seconds()
                        count_delta /= time_diff
                        if "data" in counter:
                            count_delta = self.convert_to_gbps(count_delta)
                        count_delta = round(count_delta, 2)
                        if int(count_delta) == count_delta:
                            count_delta = int(count_delta)
                    else:
                        count_delta = 0
                    row[card_idx] = f"{count}({count_delta})"
                counters.append(row)
            self.prev_counter_data = curr_counter_data

            space = 5 + max(len(counter)
                            for counter in ROCESTAT_HW_METRICS_DESC)
            print_format = f"%{space}s " + \
                " ".join(["%25s" for _ in range(len(header)-1)])
            if len(header) > 1:
                print(print_format % tuple(header))
                for counter in counters:
                    print(print_format % tuple(counter))
            print()

        # Per lane stats
        if lane_stats_flag or (lane_stats_flag == hw_stats_flag):
            lane_stats = {}
            for metric in ROCESTAT_PER_LANE_METRICS:
                idx = 0
                try:
                    val = group[metric].netValues
                except IndexError:
                    idx += 1
                    continue

                for elem in val:
                    lane = elem[1]
                    value = elem[2]

                    if lane not in lane_stats:
                        lane_stats[lane] = {
                            m: 0 for m in ROCESTAT_PER_LANE_METRICS}
                    lane_stats[lane][metric] = value

            if lane_stats:
                print("%10s %20s %12s %12s %12s %12s %12s %12s" % (
                    "Netdev", "Queue", "tx_bytes", "rx_bytes", "tx/s(Gbps)", "rx/s(Gbps)", "pause", "pause_delta"))
                for lane, metrics in lane_stats.items():
                    netdev, queue = self.split_lane(lane)
                    tx_bw_gbps = self.convert_to_gbps(
                        metrics.get('rocestat.lane.tx_bytes_bw', 0))
                    rx_bw_gbps = self.convert_to_gbps(
                        metrics.get('rocestat.lane.rx_bytes_bw', 0))
                    print("%10s %20s %12s %12s %12s %12s %12s %12s" % (
                        netdev, queue,
                        metrics.get('tx_bytes', 0),
                        metrics.get('rx_bytes', 0),
                        tx_bw_gbps,
                        rx_bw_gbps,
                        metrics.get('rx_pause', 0),
                        metrics.get('rx_pause_delta', 0)
                    ))
        print()

    def split_lane(self, lane):
        parts = lane.split('_')
        if len(parts) == 2:
            netdev = parts[0]
            queue_id = parts[1].replace("lane", "")

            queue_name = {
                "0": "Default(TCP) - 0",
                "1": "Normal Large - 1",
                "2": "VIP Small - 2",
                "3": "VIP Large - 3",
                "4": "Normal Small - 4",
                "5": "VIP Small - 5",
                "6": "Unused - 6",
                "7": "CNP - 7"
            }.get(queue_id, f"Queue - {queue_id}")

            return netdev, queue_name
        return lane, "Unknown Queue"

    def convert_to_gbps(self, value):
        if value == 0:
            return 0
        return value / 1000000000


if __name__ == '__main__':
    try:
        opts = RoceStatOptions()
        mngr = pmcc.MetricGroupManager.builder(opts, sys.argv)
        RoceStatOptions.context = mngr.type
        missing = mngr.checkMissingMetrics(
            ROCESTAT_HW_METRICS + ROCESTAT_PER_LANE_METRICS)
        if missing is not None:
            sys.stderr.write(
                'Error: not all required metrics are available\nMissing %s\n' % missing)
            sys.exit(1)
        mngr["rocestat"] = ROCESTAT_HW_METRICS + ROCESTAT_PER_LANE_METRICS
        mngr.printer = ROCeStatReport(opts.samples)
        sts = mngr.run()
        sys.exit(sts)

    except pmapi.pmErr as error:
        sys.stderr.write("%s %s\n" % (error.progname(), error.message()))
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env pmpython
#
# Copyright (C) 2023-2024 Nathan Scott <nathans@debian.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# pylint: disable=superfluous-parens
# pylint: disable=invalid-name, line-too-long, no-self-use
# pylint: disable=too-many-boolean-expressions, too-many-statements
# pylint: disable=too-many-instance-attributes, too-many-locals
# pylint: disable=too-many-branches, too-many-nested-blocks, too-many-arguments
# pylint: disable=broad-except

""" PCP to Apache Arrow (especially parquet format) Bridge """

# Common imports
from collections import OrderedDict
import sys
import os

# Arrow imports
import pyarrow as pa
import pyarrow.dataset as ds

# PCP Python API
from pcp import pmapi, pmconfig
from cpmapi import pmSetContextOptions
from cpmapi import PM_CONTEXT_ARCHIVE, PM_INDOM_NULL, PM_SEM_COUNTER
from cpmapi import PM_TYPE_32, PM_TYPE_64, PM_TYPE_U32, PM_TYPE_U64
from cpmapi import PM_TYPE_FLOAT, PM_TYPE_DOUBLE, PM_TYPE_STRING

# Default config
DEFAULT_CONFIG = ["./pcp2arrow.conf", "$HOME/.pcp2arrow.conf", "$HOME/.pcp/pcp2arrow.conf", "$PCP_SYSCONF_DIR/pcp2arrow.conf"]

# Defaults
CONFVER = 1

class PCP2ARROW(object):
    """ PCP to ARROW """
    def __init__(self):
        """ Construct object, prepare for command line handling """
        self.context = None
        self.daemonize = False
        self.pmconfig = pmconfig.pmConfig(self)
        self.opts = self.options()

        # Configuration directives
        self.keys = ('source', 'output',
                     'samples', 'interval', 'type', 'daemonize', 'version',
                     'type_prefer', 'limit_filter', 'limit_filter_force',
                     'live_filter', 'rank', 'invert_filter', 'names_change',
                     'speclocal', 'instances',
                     'include_labels')

        # The order of preference for options (as present):
        # 1 - command line options
        # 2 - options from configuration file(s)
        # 3 - built-in defaults defined below
        self.check = 0
        self.version = CONFVER
        self.source = "local:"
        self.output = None # For pmrep conf file compat only
        self.speclocal = None
        self.derived = None
        self.globals = 1
        self.samples = None # forever
        self.interval = pmapi.timeval(60)      # 60 sec
        self.opts.pmSetOptionInterval(str(60)) # 60 sec
        self.type = 0
        self.type_prefer = self.type
        self.ignore_incompat = 1
        self.ignore_unknown = 1
        self.names_change = 0 # ignore
        self.instances = []
        self.live_filter = 0
        self.rank = 0
        self.limit_filter = 0
        self.limit_filter_force = 0
        self.invert_filter = 0
        self.include_labels = 0
        self.interpol = 0

        # Internal
        self.outfile = None
        self.runtime = -1
        self.schema = None
        self.matrix = {}  # dict of value vectors, keyed by column name
        self.indoms = {}  # dict of dict, keyed by indom ID first, inst ID next
        self.column_map = {}  # maps (metric_idx, inst_id) -> column_name for fast lookup
        self.column_index = {}  # maps (metric_idx, inst_id) -> column_index for fast append
        self.columns = []  # list of column data lists for faster indexed access

        # Performance metrics store
        # key - metric name
        # values - 0:txt label, 1:instance(s), 2:unit/scale, 3:type,
        #          4:width, 5:pmfg item, 6:precision, 7:limit
        self.metrics = OrderedDict()
        self.pmfg = None
        self.pmfg_ts = None

    def options(self):
        """ Setup default command line argument option handling """
        opts = pmapi.pmOptions()
        opts.pmSetOptionCallback(self.option)
        opts.pmSetOverrideCallback(self.option_override)
        opts.pmSetShortOptions("a:h:LK:c:C:D:V?A:S:T:O:s:t:rRi:jJ:8:9:nZ:zo:")
        opts.pmSetShortUsage("[option...] [metricspec...]")

        opts.pmSetLongOptionHeader("General options")
        opts.pmSetLongOptionArchive()      # -a/--archive
        opts.pmSetLongOptionArchiveFolio() # --archive-folio
        opts.pmSetLongOptionContainer()    # --container
        opts.pmSetLongOptionHost()         # -h/--host
        opts.pmSetLongOptionLocalPMDA()    # -L/--local-PMDA
        opts.pmSetLongOptionSpecLocal()    # -K/--spec-local
        opts.pmSetLongOption("config", 1, "c", "FILE", "config file path")
        opts.pmSetLongOption("check", 0, "C", "", "check config and metrics and exit")
        opts.pmSetLongOption("output", 1, "o", "OUTFILE", "output file")
        opts.pmSetLongOptionDebug()        # -D/--debug
        opts.pmSetLongOptionVersion()      # -V/--version
        opts.pmSetLongOptionHelp()         # -?/--help

        opts.pmSetLongOptionHeader("Reporting options")
        opts.pmSetLongOptionAlign()        # -A/--align
        opts.pmSetLongOptionStart()        # -S/--start
        opts.pmSetLongOptionFinish()       # -T/--finish
        opts.pmSetLongOptionOrigin()       # -O/--origin
        opts.pmSetLongOptionSamples()      # -s/--samples
        opts.pmSetLongOptionInterval()     # -t/--interval
        opts.pmSetLongOptionTimeZone()     # -Z/--timezone
        opts.pmSetLongOptionHostZone()     # -z/--hostzone
        opts.pmSetLongOption("raw", 0, "r", "", "output raw counter values (no rate conversion)")
        opts.pmSetLongOption("raw-prefer", 0, "R", "", "prefer output raw counter values (no rate conversion)")
        #opts.pmSetLongOption("names-change", 1, "4", "ACTION", "update/ignore/abort on PMNS change (default: ignore)")
        opts.pmSetLongOption("instances", 1, "i", "STR", "instances to report (default: all current)")
        opts.pmSetLongOption("live-filter", 0, "j", "", "perform instance live filtering")
        opts.pmSetLongOption("rank", 1, "J", "COUNT", "limit results to COUNT highest/lowest valued instances")
        opts.pmSetLongOption("limit-filter", 1, "8", "LIMIT", "default limit for value filtering")
        opts.pmSetLongOption("limit-filter-force", 1, "9", "LIMIT", "forced limit for value filtering")
        opts.pmSetLongOption("invert-filter", 0, "n", "", "perform ranking before live filtering")
        #opts.pmSetLongOption("include-labels", 0, "m", "", "include metric label info")

        return opts

    def option_override(self, opt):
        """ Override standard PCP options """
        if opt in ('g', 'H', 'K', 'n', 'N'):
            return 1
        return 0

    def option(self, opt, optarg, _index):
        """ Perform setup for individual command line option """
        if opt == 'K':
            if not self.speclocal or not self.speclocal.startswith(";"):
                self.speclocal = ";" + optarg
            else:
                self.speclocal = self.speclocal + ";" + optarg
        elif opt == 'c':
            self.config = optarg
        elif opt == 'C':
            self.check = 1
        elif opt == 'e':
            if not self.derived or not self.derived.startswith(";"):
                self.derived = ";" + optarg
            else:
                self.derived = self.derived + ";" + optarg
        elif opt == 'o':
            if os.path.exists(optarg):
                sys.stderr.write("File %s already exists.\n" % optarg)
                sys.exit(1)
            self.outfile = optarg
        elif opt == 'r':
            self.type = 1
        elif opt == 'R':
            self.type_prefer = 1
        elif opt == '4':
            if optarg == 'ignore':
                self.names_change = 0
            elif optarg == 'abort':
                self.names_change = 1
            elif optarg == 'update':
                self.names_change = 2
            else:
                sys.stderr.write("Unknown names-change action '%s' specified.\n" % optarg)
                sys.exit(1)
        elif opt == 'i':
            self.instances = self.instances + self.pmconfig.parse_instances(optarg)
        elif opt == 'j':
            self.live_filter = 1
        elif opt == 'J':
            self.rank = optarg
        elif opt == '8':
            self.limit_filter = optarg
        elif opt == '9':
            self.limit_filter_force = optarg
        elif opt == 'n':
            self.invert_filter = 1
        elif opt == 'm':
            self.include_labels = 1
        else:
            raise pmapi.pmUsageErr()

    def configure(self):
        """ Process command line, read configuration and verify everything """
        os.putenv('PCP_DERIVED_CONFIG', '')

        self.pmconfig.read_cmd_line()

        ctx_type, self.source = pmapi.pmContext.set_connect_options(self.opts, self.source, self.speclocal)
        self.pmfg = pmapi.fetchgroup(ctx_type, self.source)
        self.pmfg_ts = self.pmfg.extend_timeval()
        self.context = self.pmfg.get_context()
        ctx = self.context.ctx
        # For raw archive mode, use PM_MODE_FORW to process -S/-T options
        # pmFetchArchive ignores the mode but pmSetContextOptions needs a valid mode
        if self.type == 1 and ctx_type == PM_CONTEXT_ARCHIVE:
            if pmSetContextOptions(ctx, pmapi.c_api.PM_MODE_FORW, 0):
                raise pmapi.pmUsageErr()
        else:
            if pmSetContextOptions(ctx, self.opts.mode, self.opts.delta):
                raise pmapi.pmUsageErr()

        self.config = self.pmconfig.set_config_path(DEFAULT_CONFIG)
        self.pmconfig.read_options()
        self.pmconfig.prepare_metrics(pmns=True)
        self.pmconfig.set_signal_handler()

        if self.version != CONFVER:
            sys.stderr.write("Incompatible configuration file version (read v%s, need v%d).\n" % (self.version, CONFVER))
            sys.exit(1)

        self.pmconfig.validate_common_options()

        if not self.outfile:
            sys.stderr.write("No output file name given, cannot proceed.\n")
            sys.exit(1)

        self.pmconfig.validate_metrics(curr_insts=not self.live_filter)
        self.pmconfig.finalize_options()

    def execute(self):
        """ Fetch and append values """
        # Use optimized pmFetchArchive for raw mode with archives
        if self.type == 1 and self.context.type == PM_CONTEXT_ARCHIVE:
            self.execute_raw()
            return

        # Common preparations
        self.context.prepare_execute(self.opts, False, 1, self.interval)

        # Just checking
        if self.check == 1:
            return

        # Main loop
        refresh_metrics = 0
        while self.samples != 0:
            # Refresh metrics as needed
            if refresh_metrics:
                refresh_metrics = 0
                self.pmconfig.update_metrics(curr_insts=not self.live_filter)

            # Fetch values
            refresh_metrics = self.pmconfig.fetch()
            if refresh_metrics < 0:
                break

            # Append timestamp and values
            self.append(self.pmfg_ts())

            # Finally, prepare for the next round
            if self.samples and self.samples > 0:
                self.samples -= 1
            if self.context.type != PM_CONTEXT_ARCHIVE and self.samples != 0:
                self.pmconfig.pause()

    def _iter_archive_raw(self):
        """ Iterator for raw archive reading using pmFetchArchive

        Yields (result, timestamp) tuples for each sample in the archive.
        Caller must free each result with context.pmFreeResult(result).
        """
        while True:
            try:
                result = self.context.pmFetchArchive()
                # Wrap C timespec in Python pmapi.timespec for convenient float conversion
                ts = result.contents.timestamp
                timestamp = pmapi.timespec(ts.tv_sec, ts.tv_nsec)
                yield result, timestamp
            except pmapi.pmErr as pmerr:
                if pmerr.args[0] == pmapi.c_api.PM_ERR_EOL:
                    break
                raise

    def execute_raw(self):
        """ Optimized execution using pmFetchArchive for raw archive reading """
        # Just checking
        if self.check == 1:
            return

        # Get time window bounds if specified
        start_time = float(self.opts.pmGetOptionStart()) if self.opts.pmGetOptionStart() else None
        finish_time = float(self.opts.pmGetOptionFinish()) if self.opts.pmGetOptionFinish() else None

        # Build PMID mapping for requested metrics
        pmid_to_idx = {}
        for i, _ in enumerate(self.metrics):
            desc = self.pmconfig.descs[i]
            pmid_to_idx[desc.pmid] = i

        # Use iterator for raw archive reading
        samples_read = 0
        for result, timestamp in self._iter_archive_raw():
            # Check time window (use float() for comparison with start/finish)
            timestamp_sec = float(timestamp)

            # Skip samples before start time
            if start_time is not None and timestamp_sec < start_time:
                self.context.pmFreeResult(result)
                continue

            # Stop after finish time
            if finish_time is not None and timestamp_sec > finish_time:
                self.context.pmFreeResult(result)
                break

            # Process result (convert to nanoseconds for Arrow timestamp column)
            timestamp_ns = timestamp.tv_sec * 1_000_000_000 + timestamp.tv_nsec
            self._process_raw_result(result, timestamp_ns, pmid_to_idx)

            # Free result
            self.context.pmFreeResult(result)

            samples_read += 1

            if samples_read % 100 == 0:
                if self.context.pmDebug("appl0"):
                    sys.stderr.write("  Processed %d samples...\n" % samples_read)

            # Check sample limit
            if self.samples is not None and samples_read >= self.samples:
                break

        if self.context.pmDebug("appl0"):
            sys.stderr.write("Read %d samples\n" % samples_read)

    def _process_raw_result(self, result, timestamp_ns, pmid_to_idx):
        """ Process pmFetchArchive result """
        # Initialize schema on first result
        if not self.schema:
            self.create_schema()

        # Append timestamp
        self.matrix['timestamp'].append(timestamp_ns)
        current_len = len(self.matrix['timestamp'])

        # Build dict of values from result
        # Key: (metric_idx, inst) -> value
        result_values = {}

        for i in range(result.contents.numpmid):
            pmid = result.contents.get_pmid(i)

            # Only process requested metrics
            if pmid not in pmid_to_idx:
                continue

            metric_idx = pmid_to_idx[pmid]
            metric_name = list(self.metrics.keys())[metric_idx]
            desc = self.pmconfig.descs[metric_idx]
            vset = result.contents.get_vset(i)

            if vset.contents.numval < 0:
                continue

            # Extract values for each instance
            for j in range(vset.contents.numval):
                inst = result.contents.get_inst(i, j)

                # Discover and add new instances dynamically
                if desc.indom != PM_INDOM_NULL:
                    if desc.indom not in self.indoms:
                        self.indoms[desc.indom] = {}
                    if inst not in self.indoms[desc.indom]:
                        try:
                            inst_name = self.context.pmNameInDom(desc, inst)
                            self.indoms[desc.indom][inst] = inst_name
                            # Add new column
                            metricspec = metric_name + '[' + inst_name + ']'
                            self.matrix[metricspec] = [None] * (current_len - 1)
                        except Exception:
                            self.indoms[desc.indom][inst] = str(inst)
                            metricspec = metric_name + '[' + str(inst) + ']'
                            self.matrix[metricspec] = [None] * (current_len - 1)

                try:
                    vlist = result.contents.get_vlist(i, j)
                    atom = self.context.pmExtractValue(
                        vset.contents.valfmt,
                        vlist,
                        desc.type,
                        desc.type)
                    value = atom.dref(desc.type)
                    result_values[(metric_idx, inst)] = value
                except Exception:
                    pass

        # Append values to columns (pad all columns that didn't get a value this sample)
        for i, metric in enumerate(self.metrics):
            desc = self.pmconfig.descs[i]

            if desc.indom == PM_INDOM_NULL:
                # Singular metric
                value = result_values.get((i, PM_INDOM_NULL))
                self.matrix[metric].append(value)
            else:
                # Metric with instances
                if desc.indom in self.indoms:
                    indom = self.indoms[desc.indom]
                    for instid in indom.keys():
                        name = indom[instid]
                        metricspec = metric + '[' + name + ']'
                        value = result_values.get((i, instid))
                        # Ensure column exists and has correct length
                        if metricspec not in self.matrix:
                            self.matrix[metricspec] = [None] * (current_len - 1)
                        self.matrix[metricspec].append(value)

    def lookup_indom(self, desc):
        if desc.indom in self.indoms:
            return self.indoms[desc.indom]
        indom = {}
        (insts, names) = self.pmconfig.get_metric_indom(desc)
        if insts is not None:
            for (inst, name) in zip(insts, names):
                indom[inst] = name
        self.indoms[desc.indom] = indom
        return indom

    def lookup_patype(self, desc):
        """ Find the appropriate arrow type for a metric descriptor """
        if desc.sem == PM_SEM_COUNTER and not self.type:
            return pa.float64
        if desc.type == PM_TYPE_32:
            return pa.int32
        if desc.type == PM_TYPE_64:
            return pa.int64
        if desc.type == PM_TYPE_U32:
            return pa.uint32
        if desc.type == PM_TYPE_U64:
            return pa.uint64
        if desc.type == PM_TYPE_FLOAT:
            return pa.float32
        if desc.type == PM_TYPE_DOUBLE:
            return pa.float64
        if desc.type == PM_TYPE_STRING:
            return pa.string
        return None

    def create_schema(self):
        """ Define the column names and associated types (table schema)
            starting with the timestamp column then all metrics[+insts]
        """
        self.schema = [pa.field('timestamp', pa.timestamp('ns'))]
        timestamp_col = []
        self.matrix['timestamp'] = timestamp_col
        self.columns.append(timestamp_col)

        for i, metric in enumerate(self.metrics):
            desc = self.pmconfig.descs[i]
            patype = self.lookup_patype(desc)
            if patype is None:
                continue
            if desc.indom != PM_INDOM_NULL:
                indom = self.lookup_indom(desc)
                for inst in indom.keys():
                    metricspec = metric + '[' + indom[inst] + ']'
                    col_data = []
                    col_idx = len(self.columns)
                    self.column_map[(i, inst)] = metricspec
                    self.column_index[(i, inst)] = col_idx
                    self.matrix[metricspec] = col_data
                    self.columns.append(col_data)
                    field = pa.field(metricspec, patype())
                    self.schema.append(field)
            else:
                col_data = []
                self.matrix[metric] = col_data
                self.columns.append(col_data)
                field = pa.field(metric, patype())
                self.schema.append(field)

    def append(self, timestamp):
        """ Append latest results (row) onto each arrow array (columns)
        """
        # Delayed until here as only now are values guaranteed
        if not self.schema:
            self.create_schema()

        # Append to timestamp column first (column 0)
        self.columns[0].append(timestamp)
        #print('Step:', timestamp)

        # Append either value or an Arrow nul (None) to each column
        results = self.pmconfig.get_ranked_results(valid_only=True)
        # Reuse dictionary to avoid allocations
        values = {}
        for i, metric in enumerate(self.metrics):
            desc = self.pmconfig.descs[i]
            if desc.indom == PM_INDOM_NULL:
                if metric not in results:
                    value = None
                else:
                    value = results[metric][0][2]
                self.matrix[metric].append(value)
                continue

            # Populate dictionary of values indexed by inst ID
            values.clear()
            if metric in results:
                for instid, _, value in results[metric]:
                    values[instid] = value

            # Iterate all instances for the metric, use values
            indom = self.indoms[desc.indom]
            for instid in indom.keys():
                col_idx = self.column_index[(i, instid)]
                value = values.get(instid)
                self.columns[col_idx].append(value)

    def flush(self):
        """ Create the table object and flush the dataset """
        fulltable = pa.table(self.matrix, schema=pa.schema(self.schema))
        ds.write_dataset(fulltable, self.outfile, format="parquet")

if __name__ == '__main__':
    try:
        P = PCP2ARROW()
        P.configure()

        try:
            P.execute()
        except KeyboardInterrupt:  # allow interrupt to end live sampling
            if P.context.type != PM_CONTEXT_ARCHIVE:
                sys.stdout.write("Interrupted, flushing...\n")
                P.flush()
                sys.exit(0)
            sys.stdout.write("Interrupted, skipping flush.\n")
            sys.exit(1)

        P.flush()

    except pmapi.pmErr as error:
        sys.stderr.write("%s: %s" % (error.progname(), error.message()))
        sys.stderr.write("\n")
        sys.exit(1)
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except (IOError, TypeError) as error:
        sys.stderr.write("%s\n" % str(error))
        sys.exit(1)



#!/usr/bin/pmpython
#
# Copyright (C) 2025 Red Hat.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# Create a PCP archive from GuideLLM benchmark results JSON file.
#

import argparse
import json
import sys

from datetime import datetime, timezone
from pcp import pmapi, pmi
from cpmapi import (PM_COUNT_ONE, PM_LABEL_CONTEXT,
                    PM_TIME_SEC, PM_TIME_MSEC, PM_SEM_DISCRETE,
                    PM_TYPE_DOUBLE, PM_TYPE_U64, PM_TYPE_STRING,
                    PM_TEXT_PMID, PM_TEXT_ONELINE, PM_TEXT_HELP)

def register_metrics(pcp) -> pmapi.pmInDom:
    domain = 510 # PCP domain number (guidellm import)
    indom = pcp.pmiInDom(domain, 1) # benchmark iterations
    cluster = 0 # incremented for each metric group added
    item = 0 # incremented for each metric within a group

    nounit = pcp.pmiUnits(0, 0, 0, 0, 0, 0)
    unitone = pcp.pmiUnits(0, 0, 1, 0, 0, PM_COUNT_ONE)
    secunits = pcp.pmiUnits(0, 1, 0, 0, PM_TIME_SEC, 0)
    msecunits = pcp.pmiUnits(0, 1, 0, 0, PM_TIME_MSEC, 0)
    psecunits = pcp.pmiUnits(0, 1, 1, 0, PM_TIME_SEC, PM_COUNT_ONE)

    def add_metric(name, datatype, units, oneline=None, text=None) -> None:
        nonlocal item
        pmid = pcp.pmiID(domain, cluster, item)
        pcp.pmiAddMetric(name, pmid, datatype,
                     indom, PM_SEM_DISCRETE, units)
        if oneline:
            pcp.pmiPutText(PM_TEXT_PMID, PM_TEXT_ONELINE, pmid, oneline)
        if text:
            pcp.pmiPutText(PM_TEXT_PMID, PM_TEXT_HELP, pmid, text)
        item += 1

    pmns = 'guidellm.'
    add_metric(pmns + 'run_id', PM_TYPE_STRING, nounit,
               'Unique identifier spanning the entire benchmark invocation')
    add_metric(pmns + 'duration', PM_TYPE_DOUBLE, secunits,
               'Elapsed time in seconds for each GuideLLM strategy completion')
    pmns = 'guidellm.run_stats.requests_made.'
    add_metric(pmns + 'successful', PM_TYPE_U64, unitone,
               'Count of requests processed (requests-per-second as a rate)')
    add_metric(pmns + 'errored', PM_TYPE_U64, unitone,
               'Count of failed requests, e.g. invalid input or server issues')
    add_metric(pmns + 'incomplete', PM_TYPE_U64, unitone,
               'Count of incomplete requests, e.g. timeouts or interruptions')
    add_metric(pmns + 'total', PM_TYPE_U64, unitone,
               'Count of all requests started irrespective of how they finish')

    pmns = 'guidellm.run_stats.'
    add_metric(pmns + 'finalized_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'queued_time_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'scheduled_time_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'scheduled_time_sleep_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_resolve_start_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_resolve_time_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_resolve_end_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_start_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_start_time_targeted_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_time_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'worker_targeted_start_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'request_start_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'request_start_time_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'request_start_time_targeted_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'request_targeted_start_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'request_time_delay_avg', PM_TYPE_DOUBLE, secunits)
    add_metric(pmns + 'request_time_avg', PM_TYPE_DOUBLE, secunits)

    pmns = 'guidellm.requests_per_second.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, psecunits,
        'Average number of requests processed per second',
        'Indicates throughput and the systems ability to handle concurrent workloads.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, psecunits,
        'Median number of requests processed per second',
        'Indicates throughput and the systems ability to handle concurrent workloads.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, psecunits,
        'Most frequently occurring number of requests processed per second',
        'Indicates throughput and the systems ability to handle concurrent workloads.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Total number of requests per second data points')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, psecunits,
        'Minimum requests per second')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, psecunits,
        'Maximum requests per second')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of requests per second')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within requests per second')

    pmns = 'guidellm.request_latency.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, msecunits,
        'Average time taken to process requests',
        'A critical metric for evaluating the responsiveness of the system.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, msecunits,
        'Median time taken to process requests',
        'A critical metric for evaluating the responsiveness of the system.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, msecunits,
        'Most frequently occurring time taken to process requests',
        'A critical metric for evaluating the responsiveness of the system.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Total number of time taken to process request data points')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, msecunits,
        'Minimum time taken to process a single request')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, msecunits,
        'Maximum time taken to process a single request')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of time taken to process requests')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within time taken to process requests')

    pmns = 'guidellm.time_to_first_token_ms.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, msecunits,
        'Average time taken to generate the first token of the output',
        'Indicates initial response time of the model, crucial for user-\n' +
        'facing applications.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, msecunits,
        'Median time taken to generate the first token of the output',
        'Indicates initial response time of the model, crucial for user-\n' +
        'facing applications.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, msecunits,
        'Most frequently occurring time taken to generate first output token',
        'Indicates initial response time of the model, crucial for user-\n' +
        'facing applications.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Total number of time taken to generate first output token measures')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, msecunits,
        'Minimum time taken to generate first output token')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, msecunits,
        'Maximum time taken to generate first output token')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of time taken to generate first output token')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within time taken to generate first output token')

    pmns = 'guidellm.time_per_output_token_ms.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, msecunits,
        'Average time taken to generate each output token',
        'Provides a detailed view of the models token generation efficiency.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, msecunits,
        'Median time taken to generate each output token',
        'Provides a detailed view of the models token generation efficiency.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, msecunits,
        'Most frequently occurring time taken to generate each output token',
        'Provides a detailed view of the models token generation efficiency.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Count of time taken to generate each output token measures')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, msecunits,
        'Minimum time taken to generate each output token')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, msecunits,
        'Maximum time taken to generate each output token')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of time taken to generate each output token')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within time taken to generate each output token')

    pmns = 'guidellm.inter_token_latency_ms.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, msecunits,
        'Average time between generating consecutive tokens in the output',
        'Helps assess the smoothness and speed of token generation.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, msecunits,
        'Median time between generating consecutive tokens in the output',
        'Helps assess the smoothness and speed of token generation.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, msecunits,
        'Most frequently occurring time between consecutive output tokens',
        'Helps assess the smoothness and speed of token generation.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Count of time taken between consecutive output tokens measures')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, msecunits,
        'Minimum time between generating consecutive tokens in the output')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, msecunits,
        'Maximum time between generating consecutive tokens in the output')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of time between consecutive output tokens')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within time between consecutive output tokens')

    pmns = 'guidellm.output_tokens_per_second.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, psecunits,
        'Average output tokens per second across all requests',
        'Provides insights into the servers performance and efficiency\n' +
        'in generating output tokens.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, psecunits,
        'Median prompt and output tokens per second across all requests',
        'Provides insights into the servers performance and efficiency\n' +
        'in generating output tokens.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, psecunits,
        'Mode of prompt and output tokens per second across all requests',
        'Provides insights into the servers performance and efficiency\n' +
        'in generating output tokens.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Count of output tokens per second measures')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, psecunits,
        'Minimum output tokens per second')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, psecunits,
        'Maximum output tokens per second')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of output tokens per second')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within output tokens per second')

    pmns = 'guidellm.tokens_per_second.total.'
    add_metric(pmns + 'mean', PM_TYPE_DOUBLE, psecunits,
        'Average prompt and output tokens per second across all requests',
        'Provides insights into the overall performance and efficiency\n' +
        'in processing both prompt and output tokens.')
    add_metric(pmns + 'median', PM_TYPE_DOUBLE, psecunits,
        'Median prompt and output tokens per second across all requests',
        'Provides insights into the overall performance and efficiency\n' +
        'in processing both prompt and output tokens.')
    add_metric(pmns + 'mode', PM_TYPE_DOUBLE, psecunits,
        'Mode of prompt and output tokens per second across all requests',
        'Provides insights into the overall performance and efficiency\n' +
        'in processing both prompt and output tokens.')
    add_metric(pmns + 'count', PM_TYPE_U64, unitone,
        'Count of prompt and output tokens per second measures')
    add_metric(pmns + 'min', PM_TYPE_DOUBLE, psecunits,
        'Minimum prompt and output tokens per second')
    add_metric(pmns + 'max', PM_TYPE_DOUBLE, psecunits,
        'Maximum prompt and output tokens per second')
    add_metric(pmns + 'std_dev', PM_TYPE_DOUBLE, nounit,
        'Standard deviation of prompt and output tokens per second')
    add_metric(pmns + 'variance', PM_TYPE_DOUBLE, nounit,
        'Variance within prompt and output tokens per second')

    return indom

def put_metric_values(pcp, runid, values, instance) -> None:
    """ Scan through JSON data and extract values for each PCP metric;
        pmiPutValue inserts one value for one metric:inst at one time.
    """
    pmns = 'guidellm.'
    pcp.pmiPutValue(pmns + 'duration', instance, str(values['duration']))
    pcp.pmiPutValue(pmns + 'run_id', instance, str(runid))

    pmns = 'guidellm.run_stats.requests_made.'
    stats = values['run_stats']['requests_made']
    for metric in ['successful', 'errored', 'incomplete', 'total']:
        pcp.pmiPutValue(pmns + metric, instance, str(stats[metric]))

    pmns = 'guidellm.run_stats.'
    stats = values['run_stats']
    for metric in ['finalized_delay_avg', 'queued_time_avg',
                   'request_start_delay_avg',
                   'request_targeted_start_delay_avg',
                   'request_start_time_delay_avg', 'request_time_delay_avg',
                   'request_start_time_targeted_delay_avg', 'request_time_avg'
                   'scheduled_time_delay_avg',
                   'worker_resolve_start_delay_avg', 'worker_resolve_time_avg',
                   'worker_resolve_end_delay_avg', 'worker_start_delay_avg',
                   'worker_start_time_targeted_delay_avg', 'worker_time_avg',
                   'worker_targeted_start_delay_avg']:
        try:
            pcp.pmiPutValue(pmns + metric, instance, str(stats[metric]))
        except KeyError:  # numerous variations between guidellm versions
            pass

    for group in ['requests_per_second', 'request_latency',
                  'time_to_first_token_ms', 'time_per_output_token_ms',
                  'inter_token_latency_ms', 'output_tokens_per_second',
                  'tokens_per_second']:
        pmns = 'guidellm.' + group + '.total.'
        stats = values['metrics'][group]['total']
        pcp.pmiPutValue(pmns + 'mean', instance, str(stats['mean']))
        pcp.pmiPutValue(pmns + 'mode', instance, str(stats['mode']))
        pcp.pmiPutValue(pmns + 'count', instance, str(stats['count']))
        pcp.pmiPutValue(pmns + 'median', instance, str(stats['median']))
        pcp.pmiPutValue(pmns + 'min', instance, str(stats['min']))
        pcp.pmiPutValue(pmns + 'max', instance, str(stats['max']))
        pcp.pmiPutValue(pmns + 'std_dev', instance, str(stats['std_dev']))
        pcp.pmiPutValue(pmns + 'variance', instance, str(stats['variance']))


def timestamp(timestring: str) -> datetime:
    return datetime.fromtimestamp(timestring, timezone.utc)

insts = {}    # hash instance names to instance IDs
instids = {}  # hash instance IDs to instance names

def getinstid(instance: str) -> int:
    try:
        return insts[instname]
    except KeyError:
        pass # fallthrough
    key = hash(instance) & 0x7FFFFFFF
    if instids.get(key) is None:
        instids[key] = instance
        return key
    return getinstid('_' + instance)

parser = argparse.ArgumentParser()
parser.add_argument("-v", "--verbose", help="Enable verbose progress diagnostics",
                    action="store_true")
parser.add_argument("-a", "--archive", help="Performance Co-Pilot archive output (timestamp if not specified)")
parser.add_argument("-H", "--hostname", help="Performance Co-Pilot archive hostname (run_id if not specified)")
parser.add_argument("results", help="JSON results from GuideLLM or Model Furnace")
args = parser.parse_args()

with open(args.results) as json_data:
    try:
        payload = json.load(json_data)
    except: # pylint: disable=bare-except
        print('Failed to load JSON document from:', args.results)
        sys.exit(1)
    run_id = None

    # check if the results are wrapped with model furnace metadata
    try:
        result = payload['report'] # may generate a KeyError
    except KeyError:
        result = payload # default GuideLLM format

    for benchmark in result['benchmarks']:
        if not run_id:
            run_id = benchmark['run_id']
            if args.verbose:
                print('Run', run_id)
            # Create a new PCP archive
            if not args.hostname:
                args.hostname = run_id
            if not args.archive:
                args.archive = timestamp(benchmark['run_stats']['start_time'])
                args.archive = args.archive.strftime('%Y%m%d.%H.%M')
            log = pmi.pmiLogImport(args.archive)
            log.pmiSetHostname(args.hostname)
            log.pmiSetTimezone('UTC')
            log.pmiPutLabel(PM_LABEL_CONTEXT, 0, 0, 'guidellm_run_id', run_id)
            # extract some Model Furnace metadata as labels, when present
            for label in ['model', 'inference_server', 'accelerator_type']:
                value = payload.get(label)
                if value:
                    log.pmiPutLabel(PM_LABEL_CONTEXT, 0, 0, label, value)
            ids = register_metrics(log)

        instname = benchmark['id_']
        instid = getinstid(instname)
        if args.verbose:
            print('[%d] %s' % (instid, instname))

        log.pmiAddInstance(ids, instname, instid)
        log.pmiPutLabel(PM_LABEL_CONTEXT, 0, 0, 'guidellm_id', instname)

        put_metric_values(log, run_id, benchmark, instname)
        log.pmiWrite(timestamp(benchmark['run_stats']['start_time']))

        put_metric_values(log, run_id, benchmark, instname)
        log.pmiWrite(timestamp(benchmark['run_stats']['end_time']))

        log.pmiPutMark() # end of benchmark iteration

    if args.verbose:
        print("Writing archive:", args.archive)
    del log # flush the archive to persistent storage



#!/usr/bin/env pmpython
#
# Copyright (C) 2024 Lauren Chilton <lchilton@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# pylint: disable=line-too-long
#
""" PCP to OPENMETRICS Bridge """

# Common imports
from collections import OrderedDict
import errno
import time
import sys

# Our imports
import requests
import os
import cpmapi

# PCP Python PMAPI
from pcp import pmapi, pmconfig
from cpmapi import PM_CONTEXT_ARCHIVE, PM_INDOM_NULL, PM_TIME_SEC

# Default config
DEFAULT_CONFIG = ["./pcp2openmetrics.conf", "$HOME/.pcp2openmetrics.conf", "$HOME/.pcp/pcp2openmetrics.conf", "$PCP_SYSCONF_DIR/pcp2openmetrics.conf"]

# Defaults
CONFVER = 1
INDENT = 2
TIMEFMT = "%Y-%m-%d %H:%M:%S"
TIMEOUT = 2.5 # seconds
os.environ["TZ"] = "UTC"
time.tzset()

class PCP2OPENMETRICS(object):
    """ PCP to OPENMETRICS """
    def __init__(self):
        """ Construct object, prepare for command line handling """

        self.context = None
        self.daemonize = 0
        self.pmconfig = pmconfig.pmConfig(self)
        self.opts = self.options()

        # Configuration directives
        self.keys = ('source', 'output', 'derived', 'globals',
                     'samples', 'interval', 'precision', 'daemonize',
                     'timefmt', 'everything',
                     'count_scale', 'space_scale', 'time_scale', 'version',
                     'count_scale_force', 'space_scale_force', 'time_scale_force',
                     'precision_force', 'limit_filter', 'limit_filter_force',
                     'live_filter', 'rank', 'invert_filter', 'predicate', 'names_change',
                     'speclocal', 'instances', 'ignore_incompat', 'ignore_unknown',
                     'omit_flat', 'include_labels', 'url', 'http_user', 'http_pass',
                     'http_timeout', 'no_comment')

        # Ignored for pmrep(1) compatibility
        self.keys_ignore = (
                     'timestamp','header', 'unitinfo', 'colxrow', 'separate_header', 'fixed_header',
                     'delay', 'width', 'delimiter', 'extcsv', 'width_force', 'csv_unitinfo',
                     'extheader', 'repeat_header', 'interpol',
                     'dynamic_header', 'overall_rank', 'overall_rank_alt', 'sort_metric',
                     'instinfo', 'include_texts', 'type', 'type_prefer')

        # The order of preference for options (as present):
        # 1 - command line options
        # 2 - options from configuration file(s)
        # 3 - built-in defaults defined below
        self.check = 0
        self.version = CONFVER
        self.source = "local:"
        self.output = None # For pmrep conf file compat only
        self.speclocal = None
        self.derived = None
        self.globals = 1
        self.samples = None # forever
        self.interval = pmapi.timeval(10)      # 10 sec
        self.opts.pmSetOptionInterval(str(10)) # 10 sec
        self.delay = 0
        self.type = 1
        self.type_prefer = self.type
        self.ignore_incompat = 0
        self.ignore_unknown = 0
        self.names_change = 0 # ignore
        self.instances = []
        self.live_filter = 0
        self.rank = 0
        self.limit_filter = 0
        self.limit_filter_force = 0
        self.invert_filter = 0
        self.predicate = None
        self.omit_flat = 0
        self.include_labels = 0
        self.precision = 3 # .3f
        self.precision_force = None
        self.timefmt = TIMEFMT
        self.interpol = 1
        self.count_scale = None
        self.count_scale_force = None
        self.space_scale = None
        self.space_scale_force = None
        self.time_scale = None
        self.time_scale_force = None

        # Not in pcp2openmetrics.conf, won't overwrite
        self.outfile = None

        self.everything = 0
        self.url = None
        self.http_user = None
        self.http_pass = None
        self.http_timeout = TIMEOUT
        self.no_comment = False
        self.headers = []

        # Internal
        self.runtime = -1

        self.data = None
        self.prev_ts = None
        self.writer = None

        # Performance metrics store
        # key - metric name
        # values - 0:txt label, 1:instance(s), 2:unit/scale, 3:type,
        #          4:width, 5:pmfg item, 6:precision, 7:limit
        self.metrics = OrderedDict()
        self.pmfg = None
        self.pmfg_ts = None

        # Read configuration and prepare to connect
        self.config = self.pmconfig.set_config_path(DEFAULT_CONFIG)
        self.pmconfig.read_options()
        self.pmconfig.read_cmd_line()
        self.pmconfig.prepare_metrics()
        self.pmconfig.set_signal_handler()

    def options(self):
        """ Setup default command line argument option handling """
        opts = pmapi.pmOptions()
        opts.pmSetOptionCallback(self.option)
        opts.pmSetOverrideCallback(self.option_override)
        opts.pmSetShortOptions("a:h:LK:c:Ce:D:V?GA:S:T:O:s:t:Ii:jJ:4:58:9:nN:vmP:0:q:b:y:Q:B:Y:F:f:Z:zXo:p:U:u:x")
        opts.pmSetShortUsage("[option...] metricspec [...]")

        opts.pmSetLongOptionHeader("General options")
        opts.pmSetLongOptionArchive()      # -a/--archive
        opts.pmSetLongOptionArchiveFolio() # --archive-folio
        opts.pmSetLongOptionContainer()    # --container
        opts.pmSetLongOptionHost()         # -h/--host
        opts.pmSetLongOptionLocalPMDA()    # -L/--local-PMDA
        opts.pmSetLongOptionSpecLocal()    # -K/--spec-local
        opts.pmSetLongOption("config", 1, "c", "FILE", "config file path")
        opts.pmSetLongOption("check", 0, "C", "", "check config and metrics and exit")
        opts.pmSetLongOption("output-file", 1, "F", "OUTFILE", "output file")
        opts.pmSetLongOption("derived", 1, "e", "FILE|DFNT", "derived metrics definitions")
        opts.pmSetLongOption("daemonize", 0, "", "", "daemonize on startup")
        opts.pmSetLongOptionDebug()        # -D/--debug
        opts.pmSetLongOptionVersion()      # -V/--version
        opts.pmSetLongOptionHelp()         # -?/--help

        opts.pmSetLongOptionHeader("Reporting options")
        opts.pmSetLongOption("no-globals", 0, "G", "", "omit global metrics")
        opts.pmSetLongOptionAlign()        # -A/--align
        opts.pmSetLongOptionStart()        # -S/--start
        opts.pmSetLongOptionFinish()       # -T/--finish
        opts.pmSetLongOptionOrigin()       # -O/--origin
        opts.pmSetLongOptionSamples()      # -s/--samples
        opts.pmSetLongOptionInterval()     # -t/--interval
        opts.pmSetLongOptionTimeZone()     # -Z/--timezone
        opts.pmSetLongOptionHostZone()     # -z/--hostzone
        opts.pmSetLongOption("ignore-incompat", 0, "I", "", "ignore incompatible instances (default: abort)")
        opts.pmSetLongOption("ignore-unknown", 0, "5", "", "ignore unknown metrics (default: abort)")
        opts.pmSetLongOption("names-change", 1, "4", "ACTION", "update/ignore/abort on PMNS change (default: ignore)")
        opts.pmSetLongOption("instances", 1, "i", "STR", "instances to report (default: all current)")
        opts.pmSetLongOption("live-filter", 0, "j", "", "perform instance live filtering")
        opts.pmSetLongOption("rank", 1, "J", "COUNT", "limit results to COUNT highest/lowest valued instances")
        opts.pmSetLongOption("limit-filter", 1, "8", "LIMIT", "default limit for value filtering")
        opts.pmSetLongOption("limit-filter-force", 1, "9", "LIMIT", "forced limit for value filtering")
        opts.pmSetLongOption("invert-filter", 0, "n", "", "perform ranking before live filtering")
        opts.pmSetLongOption("predicate", 1, "N", "METRIC", "set predicate filter reference metric")
        opts.pmSetLongOption("omit-flat", 0, "v", "", "omit single-valued metrics")
        opts.pmSetLongOption("include-labels", 0, "m", "", "include metric label info")
        opts.pmSetLongOption("timestamp-format", 1, "f", "STR", "strftime string for timestamp format")
        opts.pmSetLongOption("precision", 1, "P", "N", "prefer N digits after decimal separator (default: 3)")
        opts.pmSetLongOption("precision-force", 1, "0", "N", "force N digits after decimal separator")
        opts.pmSetLongOption("count-scale", 1, "q", "SCALE", "default count unit")
        opts.pmSetLongOption("count-scale-force", 1, "Q", "SCALE", "forced count unit")
        opts.pmSetLongOption("space-scale", 1, "b", "SCALE", "default space unit")
        opts.pmSetLongOption("space-scale-force", 1, "B", "SCALE", "forced space unit")
        opts.pmSetLongOption("time-scale", 1, "y", "SCALE", "default time unit")
        opts.pmSetLongOption("time-scale-force", 1, "Y", "SCALE", "forced time unit")

        opts.pmSetLongOption("with-everything", 0, "X", "", "write everything, incl. internal IDs")
        opts.pmSetLongOption("no-comment", 0, "x", "", "omit comment lines")

        opts.pmSetLongOption("url", 1, "u", "URL", "URL of endpoint to receive HTTP POST")
        opts.pmSetLongOption("http-timeout", 1, "o", "SECONDS", "timeout when sending HTTP POST")
        opts.pmSetLongOption("http-pass", 1, "p", "PASSWORD", "password for endpoint")
        opts.pmSetLongOption("http-user", 1, "U", "USERNAME", "username for endpoint")

        return opts

    def option_override(self, opt):
        """ Override standard PCP options """
        if opt in ('g', 'H', 'K', 'n', 'N', 'p'):
            return 1
        return 0

    def option(self, opt, optarg, _index):
        """ Perform setup for individual command line option """
        if opt == 'daemonize':
            self.daemonize = 1
        elif opt == 'K':
            if not self.speclocal or not self.speclocal.startswith(";"):
                self.speclocal = ";" + optarg
            else:
                self.speclocal = self.speclocal + ";" + optarg
        elif opt == 'c':
            self.config = optarg
        elif opt == 'C':
            self.check = 1
        elif opt == 'F':
            if os.path.exists(optarg):
                sys.stderr.write("File %s already exists.\n" % optarg)
                sys.exit(1)
            self.outfile = optarg
        elif opt == 'e':
            if not self.derived or not self.derived.startswith(";"):
                self.derived = ";" + optarg
            else:
                self.derived = self.derived + ";" + optarg
        elif opt == 'G':
            self.globals = 0
        elif opt == 'I':
            self.ignore_incompat = 1
        elif opt == '5':
            self.ignore_unknown = 1
        elif opt == '4':
            if optarg == 'ignore':
                self.names_change = 0
            elif optarg == 'abort':
                self.names_change = 1
            elif optarg == 'update':
                self.names_change = 2
            else:
                sys.stderr.write("Unknown names-change action '%s' specified.\n" % optarg)
                sys.exit(1)
        elif opt == 'i':
            self.instances = self.instances + self.pmconfig.parse_instances(optarg)
        elif opt == 'j':
            self.live_filter = 1
        elif opt == 'J':
            self.rank = optarg
        elif opt == '8':
            self.limit_filter = optarg
        elif opt == '9':
            self.limit_filter_force = optarg
        elif opt == 'n':
            self.invert_filter = 1
        elif opt == 'N':
            self.predicate = optarg
        elif opt == 'v':
            self.omit_flat = 1
        elif opt == 'm':
            self.include_labels = 1
        elif opt == 'P':
            self.precision = optarg
        elif opt == '0':
            self.precision_force = optarg
        elif opt == 'f':
            self.timefmt = optarg
        elif opt == 'q':
            self.count_scale = optarg
        elif opt == 'Q':
            self.count_scale_force = optarg
        elif opt == 'b':
            self.space_scale = optarg
        elif opt == 'B':
            self.space_scale_force = optarg
        elif opt == 'y':
            self.time_scale = optarg
        elif opt == 'Y':
            self.time_scale_force = optarg
        elif opt == 'X':
            self.everything = 1
        elif opt == 'x':
            self.no_comment = True
        elif opt == 'u':
            self.url = optarg
        elif opt == 'o':
            self.http_timeout = float(optarg)
        elif opt == 'U':
            self.http_user = optarg
        elif opt == 'P':
            self.http_pass = optarg
        else:
            raise pmapi.pmUsageErr()

    def connect(self):
        """ Establish PMAPI context """
        context, self.source = pmapi.pmContext.set_connect_options(self.opts, self.source, self.speclocal)

        self.pmfg = pmapi.fetchgroup(context, self.source)
        self.pmfg_ts = self.pmfg.extend_timeval()
        self.context = self.pmfg.get_context()

        if pmapi.c_api.pmSetContextOptions(self.context.ctx, self.opts.mode, self.opts.delta):
            raise pmapi.pmUsageErr()

    def validate_config(self):
        """ Validate configuration options """
        if self.version != CONFVER:
            sys.stderr.write("Incompatible configuration file version (read v%s, need v%d).\n" % (self.version, CONFVER))
            sys.exit(1)

        self.pmconfig.validate_common_options()

        self.pmconfig.validate_metrics(curr_insts=not self.live_filter)
        self.pmconfig.finalize_options()

    def execute(self):
        """ Fetch and report """
        # Debug
        if self.context.pmDebug("appl1"):
            sys.stdout.write("Known config file keywords: " + str(self.keys) + "\n")
            sys.stdout.write("Known metric spec keywords: " + str(self.pmconfig.metricspec) + "\n")

        # Set delay mode, interpolation
        if self.context.type != PM_CONTEXT_ARCHIVE:
            self.delay = 1
            self.interpol = 1

        # Common preparations
        self.context.prepare_execute(self.opts, False, self.interpol, self.interval)

        # Just checking
        if self.check == 1:
            return

        # Daemonize when requested
        if self.daemonize == 1:
            self.opts.daemonize()

        # Align poll interval to host clock
        if self.context.type != PM_CONTEXT_ARCHIVE and self.opts.pmGetOptionAlignment():
            align = float(self.opts.pmGetOptionAlignment()) - (time.time() % float(self.opts.pmGetOptionAlignment()))
            time.sleep(align)

        # Main loop
        refresh_metrics = 0
        while self.samples != 0:
            # Refresh metrics as needed
            if refresh_metrics:
                refresh_metrics = 0
                self.pmconfig.update_metrics(curr_insts=not self.live_filter)

            # Fetch values
            refresh_metrics = self.pmconfig.fetch()
            if refresh_metrics < 0:
                break

            # Report and prepare for the next round
            self.report(self.pmfg_ts())
            if self.samples and self.samples > 0:
                self.samples -= 1
            if self.delay and self.interpol and self.samples != 0:
                self.pmconfig.pause()

        # Allow to flush buffered values / say goodbye
        self.report(None)

    def report(self, tstamp):
        """ Report metric values """
        if tstamp is not None:
            tstamp = tstamp.strftime(self.timefmt)

        self.write_openmetrics(tstamp)

    def write_openmetrics(self, timestamp):
        """ Write results in openmetrics format """
        if timestamp is None:
            # Silent goodbye, close in finalize()
            return

        ts = self.context.datetime_to_secs(self.pmfg_ts(), PM_TIME_SEC)

        if self.prev_ts is None:
            self.prev_ts = ts

        if not self.writer and not self.url:
            if self.outfile is None:
                self.writer = sys.stdout
            else:
                self.writer = open(self.outfile, 'wt')

        def get_type_string(desc):
            """ Get metric type as string """
            if desc.contents.type == pmapi.c_api.PM_TYPE_32:
                mtype = "32"
            elif desc.contents.type == pmapi.c_api.PM_TYPE_U32:
                mtype = "u32"
            elif desc.contents.type == pmapi.c_api.PM_TYPE_64:
                mtype = "64"
            elif desc.contents.type == pmapi.c_api.PM_TYPE_U64:
                mtype = "u64"
            elif desc.contents.type == pmapi.c_api.PM_TYPE_FLOAT:
                mtype = "float"
            elif desc.contents.type == pmapi.c_api.PM_TYPE_DOUBLE:
                mtype = "float"
            elif desc.contents.type == pmapi.c_api.PM_TYPE_STRING:
                mtype = "string"
            else:
                mtype = "unknown"
            return mtype

        def openmetrics_name(metric):
            """ pcp.io metric name to openmetrics.io name conventions """
            return metric.replace('.','_')

        def openmetrics_type(desc):
            """ convert pcp.io metric metadata to openmetrics.io TYPE """
            if desc.sem == cpmapi.PM_SEM_COUNTER:
                return 'counter'
            return 'gauge'

        def openmetrics_labels(inst, name, desc, labels):
            """ filter pcp.io labels here; pick out the labels needed """
            result = ''
            if desc.indom != PM_INDOM_NULL:
                labels[1].update(instname=name, instid=inst)
            new_dict = {}
            new_dict['semantics'] = self.context.pmSemStr(desc.contents.sem)
            new_dict['type'] = get_type_string(desc)
            if desc.indom != PM_INDOM_NULL:
                new_dict['instname'] = name
                new_dict['instid'] = inst
            for key in labels:
                new_dict.update(labels[key])
            if self.everything:
                subset = list(new_dict.keys())
            else:
                subset = ['domainname', 'test', 'groupid', 'hostname', 'machineid', 'userid', 'instname', 'instid', 'agent', 'device_type', 'indom_name']
            for i, key in enumerate(subset):
                if key not in new_dict:
                    continue
                if i != 0:
                    result += ','
                result += '%s="%s"' % (key, new_dict[key])
            return '{' + result + '}'

        results = self.pmconfig.get_ranked_results(valid_only=True)

        body = ''
        for metric in results:
            #variable declaration
            i = list(self.metrics.keys()).index(metric)
            desc = self.pmconfig.descs[i]
            context = self.pmfg.get_context()
            pmid = context.pmLookupName(metric)
            labels = context.pmLookupLabels(pmid[0])
            semantics = self.context.pmSemStr(desc.contents.sem)
            units = desc.contents.units
            pmIDStr = context.pmIDStr(pmid[0])
            pmIndomStr = context.pmInDomStr(desc)
            help_dict = {}
            help_dict[metric] = context.pmLookupText(pmid[0])

            if metric not in self.headers:
                if self.no_comment is False:
                    body += '# PCP5 %s %s %s %s %s %s\n' % (openmetrics_name(metric), pmIDStr, get_type_string(desc), pmIndomStr, semantics, units)
                body += '# TYPE %s %s\n' % (openmetrics_name(metric), openmetrics_type(desc))
                body += '# HELP %s %s\n' % (openmetrics_name(metric), help_dict[metric])
                self.headers.append(metric)

            for inst, name, value in results[metric]:
                if isinstance(value, float):
                    fmt = "." + str(self.metrics[metric][6]) + "f"
                    value = format(value, fmt)
                elif isinstance (value, int):
                    fmt = "d"
                    value = format(value, fmt)
                else:
                    str(value)

                if openmetrics_type(desc) == "counter":
                    openmetrics_name_end = openmetrics_name(metric) + "_total"
                else:
                    openmetrics_name_end = openmetrics_name(metric)

                if self.context.type == PM_CONTEXT_ARCHIVE:
                    body += '%s%s %s %s\n' % (openmetrics_name_end, openmetrics_labels(inst, name, desc, labels), value, ts)
                else:
                    body += '%s%s %s\n' % (openmetrics_name_end, openmetrics_labels(inst, name, desc, labels), value)

        if self.url:
            auth = None
            if self.http_user and self.http_pass:
                auth = requests.auth.HTTPBasicAuth(self.http_user, self.http_pass)
            try:
                timeout = self.http_timeout
                headers = {'Content-Type': 'application/openmetrics-text'}
                res = requests.post(self.url, data=body, auth=auth, headers=headers, timeout=timeout)
                if res.status_code > 299:
                    msg = "Cannot send metrics: HTTP code %s\n" % str(res.status_code)
                    sys.stderr.write(msg)
            except requests.exceptions.ConnectionError as post_error:
                msg = "Cannot connect to server at %s: %s\n" % (self.url, str(post_error))
                sys.stderr.write(msg)
        elif self.outfile:
            self.writer.write(body)
        else:
            sys.stdout.write(body)

    def finalize(self):
        """ Finalize and clean up """
        if self.writer:
            try:
                self.writer.write("# EOF\n")
                self.writer.flush()
            except IOError as write_error:
                if write_error.errno != errno.EPIPE:
                    raise
            try:
                self.writer.close()
            except Exception:
                pass
            self.writer = None

if __name__ == '__main__':
    try:
        P = PCP2OPENMETRICS()
        P.connect()
        P.validate_config()
        P.execute()
        P.finalize()
    except pmapi.pmErr as error:
        sys.stderr.write("%s: %s" % (error.progname(), error.message()))
        if error.message() == "Connection refused":
            sys.stderr.write("; is pmcd running?")
        sys.stderr.write("\n")
        sys.exit(1)
    except pmapi.pmUsageErr as usage:
        usage.message()
        sys.exit(1)
    except IOError as error:
        if error.errno != errno.EPIPE:
            sys.stderr.write("%s\n" % str(error))
            sys.exit(1)
    except KeyboardInterrupt:
        sys.stdout.write("\n")
        P.finalize()



#!/usr/bin/env pmpython
#
# Copyright (C) 2025 Red Hat.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

# pylint: disable=too-many-arguments,too-many-positional-arguments
# pylint: disable=global-statement

import logging
import time
import argparse
import os
import subprocess
import json
from pathlib import Path
from datetime import datetime, timezone

import cpmapi as api
from pcp import pmapi

imported_archives = {}
minimum_start_time = 0.0
maximum_finish_time = 0.0


def base_archive_path(path: Path):
    # From a path, return canonical PMAPI archive context string
    # (i.e. removing optional compression and .meta suffixes)
    while len(path.suffixes) != 0:
        if '.meta' not in path.suffixes:  # stripped too far, bail out now
            break
        path = path.with_suffix("")  # strip .meta, optional .gz/.xz/... from path
    return str(path)


def archive_unchanged(archive_path: str, archive_time: float, i: int, count: int):
    # Test whether we have fully imported this archive already
    previous_time = imported_archives.get(archive_path)
    if previous_time and previous_time == archive_time:
        logging.info("Skipping archive %s (no changes) [%d/%d]", archive_path, i, count)
        return True
    return False


def format_time(seconds: float):
    # From a floating point number of seconds since the epoch,
    # produce a time string in the format Grafana is expecting.
    string = datetime.fromtimestamp(seconds, timezone.utc).isoformat()
    return string.replace('+00:00', 'Z')


def update_time_window(begin: str, end: str, json_path: str, no_op: bool):
    # Update the JSON file to use the archive start and end time
    begin = format_time(begin)
    end = format_time(end)

    # Open and load the json file into a dictionary
    try:
        with open(json_path, "r", encoding="utf-8") as file:
            data = json.load(file)
    except (json.JSONDecodeError, FileNotFoundError) as e:
        logging.error("Error reading JSON file: %s", e)
        return

    # Update the from and to fields
    try:
        data["time"]["from"] = begin
        data["time"]["to"] = end
    except KeyError:
        return # no update required

    if no_op:
        logging.info("Would update the JSON file: %s", json_path)
    else:
        # Write the modified json back into the json file
        with open(json_path, "w", encoding="utf-8") as file:
            json.dump(data, file, indent=4)

        logging.info("Successfully updated the JSON file: %s", json_path)


def setup_time_window(path: Path, i: int, count: int, json_path: str, no_op: bool):
    global minimum_start_time, maximum_finish_time

    archive_path = base_archive_path(path)
    archive_mod_time = os.path.getmtime(path)
    if archive_unchanged(archive_path, archive_mod_time, i, count):
        return

    try:
        ctx = pmapi.pmContext(api.PM_CONTEXT_ARCHIVE, archive_path)
        ctx.pmNewZone('UTC')
        label = ctx.pmGetArchiveLabel()
    except pmapi.pmErr:
        logging.info("Skipping archive %s (no context) [%d/%d]", archive_path, i, count)
        return # .meta exists, but not a PCP archive metadata file
    start = float(label.start)
    if minimum_start_time == 0.0 or start < minimum_start_time:
        logging.info("Updating start to %s from %s [%d/%d]",
                     format_time(start), archive_path, i, count)
        minimum_start_time = start

    finish = float(ctx.pmGetArchiveEnd())
    if finish > maximum_finish_time:
        logging.info("Updating finish to %s from %s [%d/%d]",
                     format_time(finish), archive_path, i, count)
        maximum_finish_time = finish

    # update the time window in the JSON file with archive start and end times
    update_time_window(minimum_start_time, maximum_finish_time, json_path, no_op)


def import_archive(path: Path, i: int, count: int, no_op: bool, import_timeout: int, port: str, time_zone: str):
    archive_path = base_archive_path(path)
    archive_mod_time = os.path.getmtime(path)
    if archive_unchanged(archive_path, archive_mod_time, i, count):
        return

    start_dt = datetime.now()
    try:
        # mainly for testing/debugging
        # if --noop option is set, skip the importing of the archives
        if no_op:
            logging.info("Would run pmseries --load %s... [%d/%d]", archive_path, i, count)
        else:
            # initialize pmseries command, append optional import options
            command_line = ["pmseries", "--load", archive_path]
            if port is not None:
                command_line.extend(["-p", port])
            if time_zone is not None:
                command_line.extend(["-Z", time_zone])
            logging.info("Importing archive %s... [%d/%d]", archive_path, i, count)
            subprocess.run(
                    command_line,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    check=True,
                    universal_newlines=True,
                    timeout=import_timeout,
                )
    except subprocess.CalledProcessError as e:
        logging.error("Error: %s", e.stdout)
    except subprocess.TimeoutExpired:
        logging.error("Timeout")
    else:
        total_sec = (datetime.now() - start_dt).total_seconds()
        minutes, seconds = divmod(total_sec, 60)
        logging.info("Successfully imported archive in %d:%dm.", minutes, seconds)
        imported_archives[archive_path] = archive_mod_time


def poll(archives_path: str, jsonfile_path: str, no_op: bool, import_timeout: int, port: str, time_zone: str):
    logging.info("Searching for new or updated archives...")

    if not os.access(archives_path, os.R_OK):
        logging.error("Cannot read directory: %s", archives_path)
        logging.error("Please use '--security-opt label=disable' when starting the container.")
        return

    archive_paths = list(Path(archives_path).rglob("*.meta*"))
    if not archive_paths:
        logging.warning("No archives found.")
    else:
        # prepare the dashboard with an initial (quick) pass over all archives
        # because the import_archive process may be loading large data volumes
        for i, path in enumerate(archive_paths, start=1):
            setup_time_window(path, i, len(archive_paths), jsonfile_path, no_op)
        for i, path in enumerate(archive_paths, start=1):
            import_archive(path, i, len(archive_paths), no_op, import_timeout, port, time_zone)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--archives", default="/archives")
    parser.add_argument("--jsonfile",
                        default="/usr/local/var/lib/grafana/dashboards/pcp-archive-analysis.json")
    parser.add_argument("--noop", action="store_true")
    parser.add_argument("--poll-interval", type=int, default=10)
    parser.add_argument("--import-timeout", type=int, default=600)
    parser.add_argument('-p', "--port", type=str)
    parser.add_argument('-Z', "--timezone", type=str)
    args = parser.parse_args()
    dash = 'http://localhost:3000/dashboards'
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    logging.info("Starting Performance Co-Pilot archive import...")
    logging.info("Dashboard: %s (when using default instructions)", dash)

    # ensure archive timestamps for Grafana handled in UTC
    os.environ['TZ'] = 'UTC'
    time.tzset()

    try:
        while True:
            logging.info("Poll interval: %d", args.poll_interval)
            logging.info("Import timeout: %d", args.import_timeout)
            poll(args.archives, args.jsonfile, args.noop, args.import_timeout, args.port,
                 args.timezone)
            time.sleep(args.poll_interval)
    except KeyboardInterrupt:
        pass  # debugging

if __name__ == "__main__":
    main()
